import re
from typing import Union

colors = {
    'aliceblue': [0.941, 0.973, 1.0],
    'antiquewhite': [0.98, 0.922, 0.844],
    'aqua': [0, 1.0, 1.0],
    'aquamarine': [0.5, 1.0, 0.832],
    'azure': [0.941, 1.0, 1.0],
    'beige': [0.961, 0.961, 0.863],
    'bisque': [1.0, 0.895, 0.77],
    'black': [0, 0, 0],
    'blanchedalmond': [1.0, 0.922, 0.805],
    'blue': [0, 0, 1.0],
    'blueviolet': [0.543, 0.172, 0.887],
    'brown': [0.648, 0.168, 0.168],
    'burlywood': [0.871, 0.723, 0.531],
    'cadetblue': [0.375, 0.621, 0.629],
    'chartreuse': [0.5, 1.0, 0],
    'chocolate': [0.824, 0.414, 0.121],
    'coral': [1.0, 0.5, 0.316],
    'cornflowerblue': [0.395, 0.586, 0.93],
    'cornsilk': [1.0, 0.973, 0.863],
    'crimson': [0.863, 0.082, 0.238],
    'cyan': [0, 1.0, 1.0],
    'darkblue': [0, 0, 0.547],
    'darkcyan': [0, 0.547, 0.547],
    'darkgoldenrod': [0.723, 0.527, 0.047],
    'darkgray': [0.664, 0.664, 0.664],
    'darkgrey': [0.664, 0.664, 0.664],
    'darkgreen': [0, 0.395, 0],
    'darkkhaki': [0.742, 0.719, 0.422],
    'darkmagenta': [0.547, 0, 0.547],
    'darkolivegreen': [0.336, 0.422, 0.188],
    'darkorange': [1.0, 0.551, 0],
    'darkorchid': [0.602, 0.199, 0.801],
    'darkred': [0.547, 0, 0],
    'darksalmon': [0.914, 0.59, 0.48],
    'darkseagreen': [0.562, 0.738, 0.562],
    'darkslateblue': [0.285, 0.242, 0.547],
    'darkslategray': [0.188, 0.312, 0.312],
    'darkslategrey': [0.188, 0.312, 0.312],
    'darkturquoise': [0, 0.809, 0.82],
    'darkviolet': [0.582, 0, 0.828],
    'deeppink': [1.0, 0.082, 0.578],
    'deepskyblue': [0, 0.75, 1.0],
    'dimgray': [0.414, 0.414, 0.414],
    'dimgrey': [0.414, 0.414, 0.414],
    'dodgerblue': [0.121, 0.566, 1.0],
    'firebrick': [0.699, 0.137, 0.137],
    'floralwhite': [1.0, 0.98, 0.941],
    'forestgreen': [0.137, 0.547, 0.137],
    'fuchsia': [1.0, 0, 1.0],
    'gainsboro': [0.863, 0.863, 0.863],
    'ghostwhite': [0.973, 0.973, 1.0],
    'gold': [1.0, 0.844, 0],
    'goldenrod': [0.855, 0.648, 0.129],
    'gray': [0.504, 0.504, 0.504],
    'grey': [0.504, 0.504, 0.504],
    'green': [0, 0.504, 0],
    'greenyellow': [0.68, 1.0, 0.188],
    'honeydew': [0.941, 1.0, 0.941],
    'hotpink': [1.0, 0.414, 0.707],
    'indianred': [0.805, 0.363, 0.363],
    'indigo': [0.297, 0, 0.512],
    'ivory': [1.0, 1.0, 0.941],
    'khaki': [0.941, 0.902, 0.551],
    'lavender': [0.902, 0.902, 0.98],
    'lavenderblush': [1.0, 0.941, 0.961],
    'lawngreen': [0.488, 0.988, 0],
    'lemonchiffon': [1.0, 0.98, 0.805],
    'lightblue': [0.68, 0.848, 0.902],
    'lightcoral': [0.941, 0.504, 0.504],
    'lightcyan': [0.879, 1.0, 1.0],
    'lightgoldenrodyellow': [0.98, 0.98, 0.824],
    'lightgray': [0.828, 0.828, 0.828],
    'lightgrey': [0.828, 0.828, 0.828],
    'lightgreen': [0.566, 0.934, 0.566],
    'lightpink': [1.0, 0.715, 0.758],
    'lightsalmon': [1.0, 0.629, 0.48],
    'lightseagreen': [0.129, 0.699, 0.668],
    'lightskyblue': [0.531, 0.809, 0.98],
    'lightslategray': [0.469, 0.535, 0.602],
    'lightslategrey': [0.469, 0.535, 0.602],
    'lightsteelblue': [0.691, 0.77, 0.871],
    'lightyellow': [1.0, 1.0, 0.879],
    'lime': [0, 1.0, 0],
    'limegreen': [0.199, 0.805, 0.199],
    'linen': [0.98, 0.941, 0.902],
    'magenta': [1.0, 0, 1.0],
    'maroon': [0.504, 0, 0],
    'mediumaquamarine': [0.402, 0.805, 0.668],
    'mediumblue': [0, 0, 0.805],
    'mediumorchid': [0.73, 0.336, 0.828],
    'mediumpurple': [0.578, 0.441, 0.859],
    'mediumseagreen': [0.238, 0.703, 0.445],
    'mediumslateblue': [0.484, 0.41, 0.934],
    'mediumspringgreen': [0, 0.98, 0.605],
    'mediumturquoise': [0.285, 0.82, 0.801],
    'mediumvioletred': [0.781, 0.086, 0.523],
    'midnightblue': [0.102, 0.102, 0.441],
    'mintcream': [0.961, 1.0, 0.98],
    'mistyrose': [1.0, 0.895, 0.883],
    'moccasin': [1.0, 0.895, 0.711],
    'navajowhite': [1.0, 0.871, 0.68],
    'navy': [0, 0, 0.504],
    'oldlace': [0.992, 0.961, 0.902],
    'olive': [0.504, 0.504, 0],
    'olivedrab': [0.422, 0.559, 0.141],
    'orange': [1.0, 0.648, 0],
    'orangered': [1.0, 0.273, 0],
    'orchid': [0.855, 0.441, 0.84],
    'palegoldenrod': [0.934, 0.91, 0.668],
    'palegreen': [0.598, 0.984, 0.598],
    'paleturquoise': [0.688, 0.934, 0.934],
    'palevioletred': [0.859, 0.441, 0.578],
    'papayawhip': [1.0, 0.938, 0.836],
    'peachpuff': [1.0, 0.855, 0.727],
    'peru': [0.805, 0.523, 0.25],
    'pink': [1.0, 0.754, 0.797],
    'plum': [0.867, 0.629, 0.867],
    'powderblue': [0.691, 0.879, 0.902],
    'purple': [0.504, 0, 0.504],
    'rebeccapurple': [0.402, 0.203, 0.602],
    'red': [1.0, 0, 0],
    'rosybrown': [0.738, 0.562, 0.562],
    'royalblue': [0.258, 0.414, 0.883],
    'saddlebrown': [0.547, 0.273, 0.078],
    'salmon': [0.98, 0.504, 0.449],
    'sandybrown': [0.957, 0.645, 0.379],
    'seagreen': [0.184, 0.547, 0.344],
    'seashell': [1.0, 0.961, 0.934],
    'sienna': [0.629, 0.324, 0.18],
    'silver': [0.754, 0.754, 0.754],
    'skyblue': [0.531, 0.809, 0.922],
    'slateblue': [0.418, 0.355, 0.805],
    'slategray': [0.441, 0.504, 0.566],
    'slategrey': [0.441, 0.504, 0.566],
    'snow': [1.0, 0.98, 0.98],
    'springgreen': [0, 1.0, 0.5],
    'steelblue': [0.277, 0.512, 0.707],
    'tan': [0.824, 0.707, 0.551],
    'teal': [0, 0.504, 0.504],
    'thistle': [0.848, 0.75, 0.848],
    'tomato': [1.0, 0.391, 0.281],
    'turquoise': [0.254, 0.879, 0.816],
    'violet': [0.934, 0.512, 0.934],
    'wheat': [0.961, 0.871, 0.703],
    'white': [1.0, 1.0, 1.0],
    'whitesmoke': [0.961, 0.961, 0.961],
    'yellow': [1.0, 1.0, 0],
    'yellowgreen': [0.605, 0.805, 0.199]
}

ColorType = Union[int, float, str, list, tuple]
class PDFColor:
    """Class that generates a PDF color string (with function ``str()``)
    using the rules described in :func:`pdfme.color.parse_color`.

    Args:
        color (int, float, list, tuple, str, PDFColor): The color
            specification.
        stroke (bool, optional): Whether this is a color for stroke(True)
            or for fill(False). Defaults to False.
    """

    def __init__(
        self, color: Union[ColorType, 'PDFColor'], stroke: bool=False
    ) -> None:
        if isinstance(color, PDFColor):
            self.color = copy(color.color)
        else:
            self.color = parse_color(color)
        self.stroke = stroke

    def __eq__(self, color):
        if color is None: return self.color is None
        if not isinstance(color, PDFColor):
            return False
        return self.color == color.color and self.stroke == color.stroke

    def __neq__(self, color):
        if color is None: return not self.color is None
        if not isinstance(color, PDFColor):
            raise TypeError("Can't compare PDFColor with {}".format(type(color)))
        return self.color != color.color or self.stroke != color.stroke

    def __str__(self):
        if self.color is None:
            return ''
        if len(self.color) == 1:
            return '{} {}'.format(
                round(self.color[0], 3), 'G' if self.stroke else 'g'
            )
        if len(self.color) == 3:
            return '{} {} {} {}'.format(
                *[round(color, 3) for color in self.color[0:3]],
                'RG' if self.stroke else 'rg'
            )

def parse_color(color: ColorType) -> list:
    """Function to parse ``color`` into a list representing a PDF color.

    The scale of the colors is between 0 and 1, instead of 0 and 256, so all the
    numbers in ``color`` must be between 0 and 1.

    ``color`` of type int or float represents a gray color between black (0) and
    white (1).

    ``color`` of type list or tuple is a gray color if its length is 1, a rgb
    color if its length is 3, and a rgba color if its length is 4 (not yet
    supported).

    ``color`` of type str can be a hex color of the form "#aabbcc", the name
    of a color in the variable ``colors`` in file `color.py`_, or a space
    separated list of numbers, that is parsed as an rgb color, like
    the one described before in the list ``color`` type.

    Args:
        color (int, float, list, tuple, str): The color specification.

    Returns:
        list: list representing the PDF color.

    .. _color.py: https://github.com/aFelipeSP/pdfme/blob/main/pdfme/color.py
    """

    if color is None:
        return None
    if isinstance(color, (int, float)):
        return [color]
    if isinstance(color, str):
        if color is '':
            return None
        elif color in colors:
            return colors[color]
        elif color[0] == '#' and len(color) in [4,5,7,9]:
            try: int(color[1:], 16)
            except:
                raise TypeError("Couldn't parse hexagesimal color value: {}".format(color))

            n = len(color)
            if n in [4, 5]:
                return [int(color[i:1+i] + color[i:1+i], 16)/255 for i in range(1,4)]
            else:
                return [int(color[i:2+i], 16)/255 for i in range(1,7,2)]
        else:
            color = re.split(',| ', color)

    if not isinstance(color, (list, tuple)):
        raise TypeError('Invalid color value type: {}'.format(type(color)))

    if len(color) == 1:
        v = color[0]
        if isinstance(v, str):
            try:
                return [float(v)]
            except:
                raise TypeError("Couldn't parse numeric color value: {}".format(v))
        elif isinstance(v, (int, float)):
            return [v]
        else:
            raise TypeError("Invalid color value type: {}".format(type(v)))
    elif len(color) in [3,4]:
        try:
            return [float(c) for c in color[:4]]
        except:
            raise TypeError("Couldn't parse numeric color value: {}".format(color))

from .utils import copy



import re
from typing import Any, Iterable, Union

page_sizes = {
    'a5': (419.528, 595.276),
    'a4': (595.276, 841.89),
    'a3': (841.89, 1190.551),
    'b5': (498.898, 708.661),
    'b4': (708.661, 1000.63),
    'jis-b5': (515.906, 728.504),
    'jis-b4': (728.504, 1031.812),
    'letter': (612, 792),
    'legal': (612, 1008),
    'ledger': (792, 1224)
}

Number = Union[int, float]
MarginType = Union[int, float, Iterable[Number], dict]

def subs(string: str, *args: tuple, **kwargs: dict) -> bytes:
    """Function to take ``string``, format it using ``args`` and ``kwargs`` and
    encode it into bytes.

    Args:
        string (str): string to be transformed.

    Returns:
        bytes: the resulting bytes.
    """
    return string.format(*args, **kwargs).encode('latin')

def process_style(style: Union[str, dict], pdf: 'PDF'=None) -> dict:
    """Function to use a named style from the PDF instance passed, if ``style``
    is a string or ``style`` itself if this is a dict.

    Args:
        style (str, dict): a style name (str) or a style dict.
        pdf (PDF, optional): the PDF to extract the named style from.

    Returns:
        dict: a style dict.
    """
    if style is None:
        return {}
    elif isinstance(style, str):
        if pdf is None:
            return {}
        return copy(pdf.formats[style])
    elif isinstance(style, dict):
        return style
    else:
        raise Exception('style must be a str with the name of a style or dict')

def get_page_size(size: Union[Number, str, Iterable]) -> tuple:
    """Function to get tuple with the width and height of a page, from the value
    in ``size``.

    If ``size`` is a str, it should be the name of a page size: ``a5``, ``a4``,
    ``a3``, ``b5``, ``b4``, ``jis-b5``, ``jis-b4``, ``letter``, ``legal`` and
    ``ledger``.

    If ``size`` is a int, the page will be a square of size ``(int, int)``.

    If ``size`` is a list or tuple, it will be converted to a tuple.

    Args:
        size (int, float, str, iterable): the page size.

    Returns:
        tuple: tuple with the page width and height.
    """
    if isinstance(size, (int, float)):
        return (size, size)
    elif isinstance(size, str):
        return page_sizes[size]
    elif isinstance(size, (list, tuple)):
        return tuple(size)
    else:
        raise Exception('Page size must be a two numbers list or tuple, a'
            'number (for a square page) or any of the following strings: {}'
            .format(
                ', '. join('"{}"'.format(name) for name in page_sizes.keys())
            ))

def parse_margin(margin: MarginType) -> dict:
    """Function to transform ``margin`` into a dict containing keys ``top``,
    ``left``, ``bottom`` and ``right`` with the margins.

    If ``margin`` is a dict, it is returned as it is.

    If ``margin`` is a string, it will be splitted using commas or spaces, and
    each substring will be converted into a number, and after this, the list
    obtained will have the same treatment of an iterable.

    If ``margin`` is an iterable of 1 element, its value will be the margin for
    the four sides. If it has 2 elements, the first one will be the ``top`` and
    ``bottom`` margin, and the second one will be the ``left`` and ``right``
    margin. If it has 3 elements, these will be the ``top``, ``right`` and
    ``bottom`` margins, and the ``left`` margin will be the second number (the
    same as ``right``). If it has 4 elements, they will be the ``top``,
    ``right``, ``bottom`` and ``left`` margins respectively.

    Args:
        margin (str, int, float, tuple, list, dict): the margin element.

    Returns:
        dict: dict containing keys ``top``, ``left``, ``bottom`` and ``right``
        with the margins.
    """
    if isinstance(margin, dict):
        return margin

    if isinstance(margin, str):
        margin = re.split(',| ', margin)
        if len(margin) == 1:
            margin = float(margin)
        else:
            margin = [float(x) for x in margin]

    if isinstance(margin, (int, float)):
        margin = [margin] * 4

    if isinstance(margin, (list, tuple)):
        if len(margin) == 0:
            margin = [0] * 4
        elif len(margin) == 1:
            margin = margin * 4
        elif len(margin) == 2:
            margin = margin * 2
        elif len(margin) == 3:
            margin = margin + [margin[1]]
        elif len(margin) > 4:
            margin = margin[0:4]

        return {k: v for k, v in zip(['top', 'right', 'bottom', 'left'], margin)}
    else:
        raise TypeError('margin property must be of type str, int, list or dict')


def parse_style_str(style_str: str, fonts: 'PDFFonts') -> dict:
    """Function to parse a style string into a style dict.

    It parses a string with a semi-colon separeted list of the style attributes
    you want to apply (for a list of the attributes you can use in this string
    see :class:`pdfme.text.PDFText`). For the ones that are of type bool, you
    just have to include the name and it will mean they are ``True``,
    and for the rest you need to include the name, a colon, and the value of the
    attribute. In case the value is a color, it can be any of the possible
    string inputs to function :func:`pdfme.color.parse_color`.
    Empty values mean ``None``, and ``"1" == True`` and ``"0" == False`` for
    bool attributes.

    This is an example of a valid style string:

    .. code-block::

        ".b;s:10;c:1;u:0;bg:"

    Args:
        style_str (str): The string representing the text style.
        fonts (PDFFonts): If a font family is included, this is needed to check
            if it is among the fonts already added to the PDFFonts instance
            passed.

    Raises:
        ValueError: If the string format is not valid.

    Returns:
        dict: A style dict like the one described in :class:`pdfme.text.PDFText`.
    """

    style = {}
    for attrs_str in style_str.split(';'):
        attrs = attrs_str.split(':')
        if len(attrs) == 1:
            if attrs[0] == '':
                continue
            attr = attrs[0].strip()
            if not attr in ['b', 'i', 'u']:
                raise ValueError(
                    'Style elements with no paramter must be whether "b" for '
                    'bold, "i" for italics(Oblique) or "u" for underline.'
                )
            style[attr] = True
        elif len(attrs) == 2:
            attr = attrs[0].strip()
            value = attrs[1].strip()

            if attr in ['b', 'i', 'u']:
                if value == '1':
                    style[attr] = True
                elif value == '0':
                    style[attr] = False
                else:
                    raise ValueError(
                        'Style element "{}" must be 0 or 1: {}'
                        .format(attr, value)
                    )
            if attr == "f":
                if value not in fonts.fonts:
                    raise ValueError(
                        'Style element "f" must have the name of a font family'
                        ' already added.'
                    )

                style['f'] = value
            elif attr == "c":
                style['c'] = PDFColor(value)
            elif attr == "bg":
                style['bg'] = PDFColor(value)
            elif attrs[0] == "s":
                try:
                    v = float(value)
                    if int(v) == v:
                        v = int(v)
                    style['s'] = v
                except:
                    raise ValueError(
                        'Style element value for "s" is wrong:'
                        ' {}'.format(value)
                    )
            elif attrs[0] == 'r':
                try: style['r'] = float(value)
                except:
                    raise ValueError(
                        'Style element value for "r" is wrong:'
                        ' {}'.format(value)
                    )
            else:
                raise ValueError(
                    'Style elements with arguments must be '
                    '"b", "u", "i", "f", "s", "c", "r"'
                )

        else:
            raise ValueError('Invalid Style string: {}'.format(attrs_str))

    return style

def create_graphics(graphics: list) -> str:
    """Function to transform a list of graphics dicts (with lines and fill
    rectangles) into a PDF stream, ready to be added to a PDF page stream.

    Args:
        graphics (list): list of graphics dicts.

    Returns:
        str: a PDF stream containing the passed graphics.
    """
    last_fill = last_color = last_line_width = last_line_style = None
    stream = ''
    for g in graphics:
        if g['type'] == 'fill':
            if g['color'] != last_fill:
                last_fill = g['color']
                stream += ' ' + str(last_fill)

            stream += ' {} {} {} {} re F'.format(
                round(g['x'], 3), round(g['y'], 3),
                round(g['width'], 3), round(g['height'], 3)
            )

        if g['type'] == 'line':
            if g['color'] != last_color:
                last_color = g['color']
                stream += ' ' + str(last_color)

            if g['width'] != last_line_width:
                last_line_width = g['width']
                stream += ' {} w'.format(round(g['width'], 3))

            if g['style'] == 'dashed':
                line_style = ' 0 J [{} {}] 0 d'.format(round(g['width']*3, 3),
                    round(g['width']*1.5, 3))
            elif g['style'] == 'dotted':
                line_style = ' 1 J [0 {}] {} d'.format(round(g['width']*2, 3),
                    round(g['width'], 3)*0.5)
            elif g['style'] == 'solid':
                line_style = ' 0 J [] 0 d'
            else:
                raise Exception(
                    'line style should be dotted, dashed or solid: {}'
                    .format(g['style'])
                )

            if line_style != last_line_style:
                last_line_style = line_style
                stream += line_style

            stream += ' {} {} m {} {} l S'.format(
                round(g['x1'], 3), round(g['y1'], 3),
                round(g['x2'], 3), round(g['y2'], 3),
            )

    if stream != '':
        stream = ' q' + stream + ' Q'

    return stream

def _roman_five(n, one, five):
    return one * n if n < 4 else (one + five if n == 4 else five)

def _roman_ten(n, one, five, ten):
    return _roman_five(n, one, five) if n < 5 else ((five if n < 9 else '')
        + _roman_five(n - 5, one, ten) if n > 5 else five)

def to_roman(n: int) -> str:
    """Function to transform ``n`` integer into a string with its corresponding
    Roman representation.

    Args:
        n (int): the number to be transformed.

    Returns:
        str: the Roman representation of the integer passed.
    """
    if not (0 < n < 4000):
        raise Exception('0 < n < 4000')
    roman = ''
    n = str(int(n))
    if len(n) > 0:
        roman = _roman_ten(int(n[-1]), 'I', 'V', 'X')
    if len(n) > 1:
        roman = _roman_ten(int(n[-2]), 'X', 'L', 'C') + roman
    if len(n) > 2:
        roman = _roman_ten(int(n[-3]), 'C', 'D', 'M') + roman
    if len(n) > 3:
        roman = _roman_ten(int(n[-4]), 'M', '', '') + roman
    return roman

def get_paragraph_stream(
    x: Number, y: Number, text_stream: str, graphics_stream: str
) -> str:
    """Function to create a paragraph stream, in position ``x`` and ``y``, using
    stream information in ``text_stream`` and ``graphics_stream``.

    Args:
        x (int, float): the x coordinate of the paragraph.
        y (int, float): the y coordinate of the paragraph.
        text_stream (str): the text stream of the paragraph.
        graphics_stream (str): the graphics stream of the paragraph.

    Returns:
        str: the whole stream of the paragraph.
    """

    stream = ''
    x, y = round(x, 3), round(y, 3)
    if graphics_stream != '':
        stream += ' q 1 0 0 1 {} {} cm{} Q'.format(x, y, graphics_stream)
    if text_stream != '':
        stream += ' BT 1 0 0 1 {} {} Tm{} ET'.format(x, y, text_stream)
    return stream

def copy(obj: Any) -> Any:
    """Function to copy objects like the ones used in this project: dicts,
    lists, PDFText, PDFTable, PDFContent, etc.


    Args:
        obj (Any): the object to be copied.

    Returns:
        Any: the copy of the object passed as argument.
    """
    if isinstance(obj, list):
        return [copy(el) for el in obj]
    elif isinstance(obj, dict):
        return {k: copy(v) for k, v in obj.items()}
    else:
        return obj

class MuiltiRange:
    def __init__(self):
        self.ranges = []

    def add(self, *range_args):
        self.ranges.append(range(*range_args))

    def __contains__(self, number: Number):
        return any(number in range_ for range_ in self.ranges)



def parse_range_string(range_str: str) -> MuiltiRange:
    """Function to convert a string of comma-separated integers and integer 
    ranges into a set of all the integers included in those.

    Args:
        range_str (str): comma-separated list of integers and integer 
            ranges.

    Returns:
        MuiltiRange: a set of integers.
    """
    multi_range = MuiltiRange()
    for part in range_str.split(','):
        range_parts = part.split(':')
        if len(range_parts) == 1:
            index = int(range_parts[0].strip())
            multi_range.add(index, index + 1)
        else:
            first = range_parts[0].strip()
            range_parts[0] = 0 if first == '' else int(first)
            if len(range_parts) > 1:
                range_parts[1] = int(range_parts[1].strip())
            if len(range_parts) > 2:
                range_parts[2] = int(range_parts[2].strip())
            multi_range.add(*range_parts)

    return multi_range

from .color import PDFColor
from .fonts import PDFFonts
from .pdf import PDF



import struct
import traceback
from io import BytesIO
from pathlib import Path
from typing import BinaryIO, Union

ImageType = Union[str, Path, BinaryIO]


class PDFImage:
    """Class that represents a PDF image.

    You can pass the location path (``str`` or ``pathlib.Path`` format) of the
    image, or pass a file-like object (``io.BytesIO``) with the image bytes, the
    extension of the image, and the image name.

    Only JPEG and PNG image formats are supported in this moment. PNG images are
    converted to JPEG, and for this Pillow library is required.

    Args:
        image (str, pathlib.Path, BytesIO): The path or file-like object of the
            image.
        extension (str, optional): If ``image`` is path-like object, this
            argument should contain the extension of the image. Options are
            [``jpg``, ``jpeg``, ``png``].
        image_name (str, optional): If ``image`` is path-like object, this
            argument should contain the name of the image. This name should be
            unique among the images added to the same PDF document.
    """
    def __init__(
        self, image: ImageType, extension: str=None, image_name: str=None
    ):
        image_bytes = None
        try:
            if isinstance(image, str):
                image_bytes = Path(image).open('rb')
                self.image_name = image
                if extension is None:
                    extension = image.rpartition('.')[-1]
            elif isinstance(image, Path):
                image_bytes = image.open('rb')
                self.image_name = str(image)
                if extension is None:
                    extension = image.suffix
            elif isinstance(image, BytesIO):
                image_bytes = image
                if image_name is None:
                    raise TypeError(
                        'when image is of type io.BytesIO, image_name must be '
                        'provided'
                    )
                self.image_name = image_name
                if extension is None:
                    raise TypeError(
                        'when image is of type io.BytesIO, extension must be '
                        'provided'
                    )
            else:
                raise TypeError(
                    'image must be of type str, pathlib.Path or io.BytesIO'
                )

            if not isinstance(extension, str):
                raise TypeError('extension type is str')

            if len(extension) > 0 and extension[0] == '.':
                extension = extension[1:]

            extension = extension.strip().lower()

            if extension in ['jpg', 'jpeg']:
                self.parse_jpg(image_bytes)
            elif extension == 'png':
                self.parse_png(image_bytes)
            else:
                raise NotImplementedError(
                    'Images of type "{}" are not yet supported'.format(extension)
                )
        finally:
            if image_bytes is not None:
                image_bytes.close()

    def parse_jpg(self, bytes_: BinaryIO) -> None:
        """Method to extract metadata from a JPEG image ``bytes_`` needed to
        embed this image in a PDF document.

        This method creates this instance's attibute ``pdf_obj``, containing
        a dict that can be added to a :class:`pdfme.base.PDFBase` instance as
        a PDF Stream object that represents this image.

        Args:
            bytes_ (BinaryIO): A file-like object containing the
                image.
        """
        try:
            while True:
                markerHigh, markerLow = struct.unpack('BB', bytes_.read(2))
                if markerHigh != 0xFF or markerLow < 0xC0:
                    raise SyntaxError('No JPEG marker found')
                elif markerLow == 0xDA: # SOS
                    raise SyntaxError('No JPEG SOF marker found')
                elif (markerLow == 0xC8 or # JPG
                    (markerLow >= 0xD0 and markerLow <= 0xD9) or # RSTx
                    (markerLow >= 0xF0 and markerLow <= 0xFD)): # JPGx
                    continue
                else:
                    data_size, = struct.unpack('>H', bytes_.read(2))
                    data = bytes_.read(data_size - 2) if data_size > 2 else b''
                    if (
                        (markerLow >= 0xC0 and markerLow <= 0xC3) or #SOF0-SOF3
                        (markerLow >= 0xC5 and markerLow <= 0xC7) or #SOF4-SOF7
                        (markerLow >= 0xC9 and markerLow <= 0xCB) or #SOF9-SOF11
                        (markerLow >= 0xCD and markerLow <= 0xCF) #SOF13-SOF15
                    ):
                        depth, h, w, layers = struct.unpack_from('>BHHB', data)

                        if layers == 3: colspace = b'/DeviceRGB'
                        elif layers == 4: colspace = b'/DeviceCMYK'
                        else: colspace = b'/DeviceGray'

                        break
        except Exception:
            traceback.print_exc()
            raise ValueError(
                "Couldn't process image: {}".format(self.image_name)
            )

        bytes_.seek(0)
        image_data = bytes_.read()
        bytes_.close()

        self.width = int(w)
        self.height = int(h)

        self.pdf_obj = {
            'Type': b'/XObject',
            'Subtype': b'/Image',
            'Width': self.width,
            'Height': self.height,
            'ColorSpace': colspace,
            'BitsPerComponent': int(depth),
            'Filter': b'/DCTDecode',
            '__skip_filter__': True,
            '__stream__': image_data
        }

    def parse_png(self, bytes_: BinaryIO) -> None:
        """Method to convert a PNG image to a JPEG image and later parse it as
        a JPEG image.

        This method creates this instance's attibute ``pdf_obj``, containing
        a dict that can be added to a :class:`pdfme.base.PDFBase` instance as
        a PDF Stream object that represents this image.

        Args:
            bytes_ (BinaryIO): A file-like object containing the
                image.
        """
        from PIL import Image  # type: ignore

        im = Image.open(bytes_).convert('RGB')
        bytes_io = BytesIO()
        im.save(bytes_io, "JPEG")
        bytes_io.seek(0)
        self.parse_jpg(bytes_io)
        bytes_io.close()
        im.close()



import re
from typing import Iterable, Union
class PDFObject:
    """A class that represents a PDF object.

    This object has a :class:`pdfme.parser.PDFRef` ``id`` attribute representing
    the id of this object inside the PDF document, and acts as a dict, so the
    user can update any property of this PDF object like you would do with a
    dict.

    Args:
        id_ (PDFRef): The id of this object inside the PDF document.
        obj (dict, optional): the dict representing the PDF object.
    """
    def __init__(self, id_: 'PDFRef', obj: dict=None) -> None:
        if not isinstance(id_, PDFRef):
            raise TypeError('id_ argument must be of type PDFRef')
        self.id = id_
        self.value = {} if obj is None else obj

    def __getitem__(self, name):
        return self.value[name]

    def __setitem__(self, name, value):
        self.value[name] = value

    def __delitem__(self, name):
        del self.value[name]

    def __contains__(self, name):
        return name in self.value

class PDFRef(int):
    """An ``int`` representing the id of a PDF object.

    This is a regular ``int`` that has an additional property called ``ref``
    with a representation of this object, to be referenced elsewhere in the PDF
    document.
    """
    def __new__(cls, id_):
        return int.__new__(cls, id_)
    @property
    def ref(self) -> bytes:
        """
        Returns:
            bytes: bytes with a representation of this object, to be referenced
            elsewhere in the PDF document.
        """
        return subs('{} 0 R', self)

ObjectType = Union[
    PDFObject, PDFRef, dict, list, tuple, set, bytes, bool, int, float, str
]

def parse_obj(obj: ObjectType) -> bytes:
    """Function to convert a python object to a bytes object representing the
    corresponding PDF object.

    Args:
        obj (PDFObject, PDFRef, dict, list, tuple, set, bytes, bool, int, float,
            str): the object to be converted to a PDF object.

    Returns:
        bytes: bytes representing the corresponding PDF object.
    """
    if isinstance(obj, PDFObject):
        return parse_obj(obj.value)
    elif isinstance(obj, PDFRef):
        return obj.ref
    elif isinstance(obj, dict):
        if '__stream__' in obj:
            return parse_stream(obj)
        else:
            return parse_dict(obj)
    elif isinstance(obj, (list, tuple, set)):
        return parse_list(obj)
    elif isinstance(obj, bytes):
        return obj
    elif isinstance(obj, bool):
        return b'true' if obj else b'false'
    elif isinstance(obj, (int, float)):
        return str(obj).encode('latin')
    elif isinstance(obj, str):
        return ('(' + re.sub(r'([()])', r'\\\1', obj) + ')').encode('latin')


def parse_dict(obj: dict) -> bytes:
    """Function to convert a python dict to a bytes object representing the
    corresponding PDF Dictionary.

    Args:
        obj (dict): the dict to be converted to a PDF Dictionary.

    Returns:
        bytes: bytes representing the corresponding PDF Dictionary.
    """
    bytes_ = b'<<'
    for key, value in obj.items():
        bytes_ += b'/' + key.encode('latin')
        ret = parse_obj(value)
        if not ret[0] in [b'/', b'(', b'<']: bytes_ += b' '
        bytes_ += ret

    return bytes_ + b'>>'

def parse_list(obj: Iterable) -> bytes:
    """Function to convert a python iterable to a bytes object representing the
    corresponding PDF Array.

    Args:
        obj (iterable): the iterable to be converted to a PDF Array.

    Returns:
        bytes: bytes representing the corresponding PDF Array.
    """
    bytes_ = b'['
    for i, value in enumerate(obj):
        ret = parse_obj(value)
        if not ret[0] in [b'/', b'(', b'<'] and i != 0: bytes_ += b' '
        bytes_ += ret

    return bytes_ + b']'

def parse_stream(obj: dict) -> bytes:
    """Function to convert a dict representing a PDF Stream object to a bytes
    object.

    A dict representing a PDF stream should have a ``'__stream__`` key
    containing the stream bytes. You don't have to include ``Length`` key in the
    dict, as it is calculated by us. The value of ``'__stream__'`` key must
    be of type ``bytes`` or a dict whose values are of type ``bytes``.
    If you include a ``Filter`` key, a encoding is automatically done in the
    stream (see :meth:`pdfme.encoders.encode_stream` function for
    supported encoders). If the contents of the stream are already encoded
    using the filter in ``Filter`` key, you can skip the encoding process
    by including the ``__skip_filter__`` key.

    Args:
        obj (dict): the dict representing a PDF stream.

    Returns:
        bytes: bytes representing the corresponding PDF Stream.
    """
    stream_ = obj.pop('__stream__')
    skip_filter = obj.pop('__skip_filter__', False)

    if isinstance(stream_, bytes):
        stream_str = stream_
    elif isinstance(stream_, dict):
        stream_str = b''.join(stream_.values())
    else:
        raise Exception(
            'streams must be bytes or a dict of bytes: ' + str(stream_)
        )

    stream = encode_stream(stream_str, obj['Filter']) \
        if 'Filter' in obj and not skip_filter else stream_str

    obj['Length'] = len(stream)
    ret = parse_dict(obj) + b'stream\n' + stream + b'\nendstream'
    obj['__stream__'] = stream_
    if skip_filter:
        obj['__skip_filter__'] = True
    return ret

from .encoders import encode_stream
from .utils import subs



from pdfme import build_pdf

from .utils import gen_text

def test_running_section_per_page():
    document = {
        "running_sections": {
            "header": {
                "x": "left", "y": 20, "height": "top", "style": {"text_align": "r"},
                "content": [{".b": "This is a header"}]
            }
        },
        'per_page': [
            {'pages': '1:1000:2', 'style': {'margin': [60, 100, 60, 60]}},
            {'pages': '0:1000:2', 'style': {'margin': [60, 60, 60, 100]}},
            {'pages': '0,4:40:2', 'running_sections': {'include': ['header']}},
            {'pages': '2', 'running_sections': {'exclude': ['header']}},
        ],
        "sections": [
            {
                "content": [gen_text(100) for _ in range(50)]
            },
        ]
    }

    with open('test_running_section_per_page.pdf', 'wb') as f:
        build_pdf(document, f)
   


import json
import re
from typing import Optional, Union
from uuid import uuid4

PARAGRAPH_DEFAULTS = {'text_align': 'l', 'line_height': 1.1, 'indent': 0}
TEXT_DEFAULTS = {'f': 'Helvetica', 'c': 0.1, 's': 11, 'r': 0, 'bg': None}

ContentType = Union[str, list, tuple, dict]
Number = Union[int, float]
class PDFState:
    """Class that represents the state of a paragraph line part.

    The state is a lower level version of the style, and is used by the other
    paragraph classes to make calculations and yield the paragraph PDF stream.

    Args:
        style (dict): the paragraph line part style.
        fonts (PDFFonts): the fonts instance with the information about
            the fonts already added to the PDF document.
    """
    def __init__(self, style: dict, fonts: 'PDFFonts') -> None:
        self.font_family = style['f']

        f_mode = ''
        if style.get('b', False):
            f_mode += 'b'
        if style.get('i', False):
            f_mode += 'i'
        if f_mode == '':
            f_mode = 'n'
        self.font_mode = 'n' if not f_mode in fonts.fonts[style['f']] else f_mode

        self.font = fonts.get_font(self.font_family, self.font_mode)

        self.size = style['s']
        self.color = PDFColor(style['c'])
        self.rise = style.get('r', 0) * self.size

    def compare(self, other: Optional['PDFState']) -> str:
        """Compares this state, with state ``other`` and returns a PDF stream
        with the differences between both states.

        Args:
            other (PDFState): the state to compare.

        Returns:
            str: a PDF stream with the differences between both states.
        """
        ret_value = ''
        if (
            other is None or self.font_family != other.font_family or
            self.font_mode != other.font_mode or self.size != other.size
        ):
            ret_value += ' /{} {} Tf'.format(self.font.ref, round(self.size, 3))
        if other is None or self.color != other.color:
            ret_value += ' ' + str(self.color)
        if other is None or self.rise != other.rise:
            ret_value += ' {} Ts'.format(round(self.rise, 3))

        return ret_value

class PDFTextLinePart:
    """This class represents a part of a paragraph line, with its own style.

    Args:
        style (dict): the style of this line part.
        fonts (PDFFonts): the fonts instance with the information about
            the fonts already added to the PDF document.
        ids (list, optional): the ids of this part.
    """
    def __init__(self, style: dict, fonts: 'PDFFonts', ids: list=None) -> None:
        self.fonts = fonts

        self.style = style
        self.state = PDFState(style, fonts)
        self.underline = style.get('u', False)
        self.background = PDFColor(style.get('bg'))
        self.ids = [] if id is None else ids
        self.width = 0
        self.words = []

        self.space_width = self.get_char_width(' ')
        self.spaces_width = 0

    def pop_word(self, index: int=None) -> Optional[str]:
        """Function to delete the last word of this part if ``index`` is None,
        and the word in the position ``index`` if it's not None.

        Args:
            index (int, optional): word index.

        Returns:
            str: if word in ``index`` could be deleted, the deleted word is
                returned, if not None is returned.
        """
        if len(self.words) > 0:
            word = self.words.pop() if index is None else self.words.pop(index)
            if word == ' ':
                self.spaces_width -= self.space_width
            else:
                self.width -= self.get_word_width(word)
            return word

    def add_word(self, word: str) -> None:
        """Function to add a word to this part.

        Args:
            word (str): the word.
        """
        self.words.append(word)
        if word == ' ':
            self.spaces_width += self.space_width
        else:
            self.width += self.get_word_width(word)

    def current_width(self, factor: Number=1) -> float:
        """Return the width of this part, according to the words added to this
        part, using ``factor`` to calculate this width of the spaces in this
        part.

        Args:
            factor (int, float, optional): to calculate this width of the spaces
                in this part.

        Returns:
            float: width of this part.
        """
        return self.width + self.spaces_width*factor

    def tentative_width(self, word: str, factor: Number=1) -> float:
        """The same as method ``current_width``, but adding the width of
        ``word``.

        Args:
            word (str): the word that could be added to this part.
            factor (int, float, optional): to calculate this width of the spaces
                in this part.

        Returns:
            float: the width of this part + the width of the word passed.
        """
        word_width = self.space_width * factor if word == ' ' else \
            self.get_word_width(word)
        return self.current_width(factor) + word_width

    def get_char_width(self, char: str) -> float:
        """The width of the character passed.

        Args:
            char (str): the character string.

        Returns:
            float: the width of the character passed.
        """
        return self.state.size * self.state.font.get_char_width(char)

    def get_word_width(self, word: str) -> float:
        """The width of the word passed.

        Args:
            char (str): the word string.

        Returns:
            float: the width of the word passed.
        """
        return self.state.size * self.state.font.get_text_width(word)

class PDFTextLine:
    """Class that represents a line of a paragraph.

    This class has the logic to add paragraph parts, and inside them add their
    words one by one, until all of the horizontal space of the paragraph has
    been used. For more information about this mechanism check the method
    :meth:`pdfme.text.PDFTextLine.add_word`.

    Args:
        fonts (PDFFonts):  to extract information about the fonts
            being used in the paragraph.
        max_width (int, float, optional): the maximum horizontal space that this
            line can use.
        text_align (str, optional): ``'l'`` for left (default), ``'c'`` for
            center, ``'r'`` for right and ``'j'`` for justified text.
        top_margin (Number, optional): if not None, this is the top margin of
            the line, added to the actual line height.
    """
    def __init__(
        self, fonts: 'PDFFonts', max_width: Number=0, text_align: str=None,
        top_margin: Number=0
    ) -> None:
        self.fonts = fonts
        self.max_width = max_width
        self.line_parts = []

        self.justify_min_factor = 0.7
        self.text_align = PARAGRAPH_DEFAULTS['text_align'] \
            if text_align is None else text_align

        self.factor = 1 if self.text_align != 'j' else self.justify_min_factor

        self.top_margin = top_margin
        self.next_line = None
        self.is_last_word_space = True
        self.firstWordAdded = False
        self.started = False

    @property
    def height(self) -> float:
        """Property that returns the line height, calculated from the vertical
        space of each part of the line.

        Returns:
            float: the line height.
        """
        top = 0
        height_ = 0
        for part in self.line_parts:
            if part.state.rise > 0 and part.state.rise > top:
                top = part.state.rise
            if part.state.size > height_:
                height_ = part.state.size

        return height_ + self.top_margin + top

    @property
    def min_width(self) -> float:
        """Property that returns the width of the line, calculated using the
        minimum value for attribute ``factor``. This attribute is used to
        increase or decrease the space character width inside a line to

        Returns:
            float: the line width.
        """
        ws = self.get_widths()
        return ws[0] + ws[1] * self.factor

    @property
    def bottom(self) -> float:
        """Property that returns the line bottom, calculated from the vertical
        space of each part of the line.

        Returns:
            float: the line bottom.
        """
        bottom = 0
        for part in self.line_parts:
            if part.state.rise < 0 and -part.state.rise > bottom:
                bottom = -part.state.rise
        return bottom

    def get_widths(self) -> tuple:
        """This function returns the widths of the line.

        Returns:
            tuple: of 2 elements, the width on the words as the first, and the
                width of the spaces as the second.
        """
        words_width = 0
        spaces_width = 0
        for part in self.line_parts:
            words_width += part.width
            spaces_width += part.spaces_width
        return words_width, spaces_width

    def add_line_part(self, style:dict=None, ids:list=None) -> PDFTextLinePart:
        """Add a new line part to this line.

        Args:
            style (dict, optional): the style of the new part.
            ids (list, optional): the ids of the new part.

        Returns:
            PDFTextLinePart: The new line part that was added.
        """
        if self.next_line is None:
            self.next_line = PDFTextLine(
                self.fonts, self.max_width, self.text_align
            )

        line_part = PDFTextLinePart(style, self.fonts, ids)
        self.next_line.line_parts.append(line_part)
        return line_part

    def add_accumulated(self) -> None:
        """Function to add the parts accumulated in the auxiliar line (
        ``next_line`` attribute) to this line.
        """
        if len(self.line_parts):
            for word in self.next_line.line_parts[0].words:
                self.line_parts[-1].add_word(word)
            self.next_line.line_parts = self.next_line.line_parts[1:]

        self.line_parts.extend(self.next_line.line_parts)
        last_part = self.line_parts[-1]
        last_part.add_word(' ')
        self.next_line.line_parts = [
            PDFTextLinePart(last_part.style, self.fonts, last_part.ids)
        ]

    def add_word(self, word:str) -> dict:
        """Function to add a word to this line.

        Args:
            word (str): The word to be added.

        Returns:
            dict: containing a ``status`` key, with one of the following values:

            * ``'added'``: The word passed was added to the auxiliar line, or
              if the word is a space the accumulated words in the auxiliar line,
              to the current line.

            * ``'ignored'``: The word passed (a space) was ignored.

            * ``'preadded'``: The word passed was added to the auxiliar line.

            * ``'finished'``: The word didn't fit in the current line, and this
              means this line is full. Because of this, a new line is created
              to put this word, and this new line is returned in the key
              ``'new_line'``.

        """
        if not self.started:
            if word.isspace():
                if self.firstWordAdded:
                    self.started = True
                    self.add_accumulated()
                    return {'status': 'added'}
                else:
                    return {'status': 'ignored'}
            else:
                self.firstWordAdded = True
                self.next_line.line_parts[-1].add_word(word)
                return {'status': 'added'}
        else:
            if word.isspace():
                if self.is_last_word_space:
                    return {'status': 'ignored'}
                else:
                    self.add_accumulated()
                    return {'status': 'added'}
            else:
                self.is_last_word_space = False
                self.next_line.line_parts[-1].add_word(word)
                if (self.min_width + self.next_line.min_width < self.max_width):
                    return {'status': 'preadded'}
                else:
                    if (
                        len(self.line_parts[-1].words) and
                        self.line_parts[-1].words[-1] == ' '
                    ):
                        self.line_parts[-1].pop_word(-1)
                    self.next_line.firstWordAdded = True
                    self.next_line.top_margin = self.bottom
                    self.next_line.next_line = PDFTextLine(
                        self.fonts, self.max_width, self.text_align
                    )
                    line_parts = self.next_line.line_parts
                    self.next_line.next_line.line_parts = line_parts
                    self.next_line.line_parts = []
                    return {
                        'status': 'finished', 'new_line': self.next_line
                    }

class PDFTextBase:
    """Class that represents a rich text paragraph to be added to a
    :class:`pdfme.pdf.PDF` instance.

    You should use :class:`pdfme.text.PDFText` instead of this class, because
    it has more functionalities.

    To create the data needed to add this paragraph to the PDF document,
    you have to call the method :meth:`pdfme.text.PDFTextBase.run`, which
    will try to add all of the dict parts in ``content`` argument list (or
    tuple) to the rectangle defined by args ``x``, ``y``, ``width`` and
    ``height``.

    Each part represents a part of the paragraph with a different style or with
    a ``var`` or a specific ``id``.

    The parts are added to this rectangle, until they are all
    inside of it, or until all of the vertical space is used and the rest of
    the parts can not be added. In these two cases method ``run``
    finishes, and the property ``finished`` will be True if all the parts
    were added, and False if the vertical space ran out.
    If ``finished`` is False, you can set a new rectangle (on a new page for
    example) and use method ``run`` again (passing the parameters of the new
    rectangle) to add the remaining parts that couldn't be added in the
    last rectangle. You can keep doing this until all of the parts are
    added and therefore property ``finished`` is True.

    By using method ``run`` the paragraph is not really added to the PDF
    object. After calling ``run``, the property ``result`` will be
    available with the information needed to be added to the PDF, at least
    the parts that fitted inside the rectangle. You have to use the
    property ``result`` to add the paragraph to the PDF object before
    using method ``run`` again (in case ``finished`` is False), because
    it will be redefined for the next rectangle after calling ``run`` again.
    You can check the ``text`` method in `PDF`_ module to see how this
    process is done.

    The other args not defined here, are explained in
    :class:`pdfme.text.PDFText`.

    Args:
        content (str, list, tuple): If this is a string, it will
            become the following:

            .. code-block:: python

                [{'style': <DEFAULT_STYLE>, 'text': <STRING>}]

            If this is a list or a tuple, its elements should be dicts with the
            following keys:

            * ``'text'``: this is the text that will be displayed with the style
              defined in ``style`` key.

            * ``'style'``: this is a style dict like the one described in
              :class:`pdfme.text.PDFText`.

            * ``'ids'``: see :class:`pdfme.text.PDFText` definition.

            * ``'var'``: see :class:`pdfme.text.PDFText` definition.

    Raises:
        TypeError: if ``content`` is not a str, list or tuple.

    .. _PDF: https://github.com/aFelipeSP/pdfme/blob/main/pdfme/pdf.py
    """
    def __init__(
        self, content: Union[str, list, tuple], width: Number, height: Number,
        x: Number=0, y: Number=0, fonts: 'PDFFonts'=None, text_align: str=None,
        line_height: Number=None, indent: Number=0, list_text: str=None,
        list_indent: Number=None, list_style: dict=None, pdf: 'PDF'=None
    ) -> None:
        self.fonts = fonts
        self.setup(x, y, width, height)
        self.indent = indent
        self.text_align = PARAGRAPH_DEFAULTS['text_align'] \
            if text_align is None else text_align
        self.line_height = PARAGRAPH_DEFAULTS['line_height'] \
            if line_height is None else line_height

        self.list_text = list_text
        self.list_indent = list_indent
        self.list_style = list_style
        self.pdf = pdf

        if isinstance(content, str):
            content = [{'style': TEXT_DEFAULTS.copy(), 'text': content}]
        if not isinstance(content, (list, tuple)):
            raise TypeError(
                'content must be of type str, list or tuple: {}'.format(content)
            )

        self.last_part_added = 0
        self.last_part_line = 0
        self.last_part = 0

        self.last_word_added = 0
        self.last_word_line = 0
        self.last_word = 0

        self.content = content
        self.finished = False
        self.is_first_line = True
        self.correct_indent = True
        self.list_setup_done = False

    @property
    def stream(self) -> str:
        """Property that returns the PDF stream generated by the method ``run``,
        with all of the graphics and the text, ready to be added to a PDF page
        stream.

        Returns:
            str: the stream.
        """
        return get_paragraph_stream(self.x, self.y, self.text, self.graphics)

    @property
    def result(self) -> dict:
        """Property that returns a dict with the result of calling method
        ``run``, and can be passed to method
        :meth:`pdfme.pdf.PDF._add_text`, to add this paragraph to that
        PDF document's page. Check method ``_add_parts`` from
        :class:`pdfme.pdf.PDF` to see how a dict like the one returned by
        this method (a paragraph part) is added to a PDF instance.

        The dict returned will have the following keys:

        * ``x`` the x coordinate.

        * ``y`` the y coordinate.

        * ``width`` of the paragraph.

        * ``height`` of the paragraph.

        * ``text_stream`` a string with the paragraphs PDF text stream.

        * ``graphics_stream`` a string with the paragraphs PDF graphics stream.

        * ``used_fonts`` a set with tuples of 2 elements, first element the
          font family, and second element the font mode.

        * ``ids`` a dict with every id extracted from the paragraph.

        Returns:
            dict: like the one described.

        .. _PDF: https://github.com/aFelipeSP/pdfme/blob/main/pdfme/pdf.py
        """

        return dict(
            x=self.x, y=self.y, width=self.width, height=self.current_height,
            text_stream=self.text, graphics_stream=self.graphics,
            used_fonts=self.used_fonts, ids=self.ids,
        )

    def get_state(self) -> dict:
        """Method to get the current state of this paragraph. This can be used
        later in method :meth:`pdfme.text.PDFText.set_state` to
        restore this state in this paragraph (like a checkpoint in a
        videogame).

        Returns:
            dict: a dict with the state of this paragraph.
        """
        return {
            'last_part': self.last_part,
            'last_word': self.last_word
        }

    def set_state(self, last_part: int=None, last_word: int=None) -> None:
        """Function to update the state of the paragraph

        The arguments of this method define the current state of this paragraph,
        and with this method you can change that state.

        Args:
            last_part (int): this is the index of the part that was being
                processed the last time method ``run`` was called.
            last_word (int): this is the index of the
                word of the last part that was added the last time method
                ``run`` was called.
        """
        self.last_part = last_part
        self.last_word = last_word

    def setup(
        self, x: Number=None, y:Number=None, width: Number=None,
        height: Number=None
    ):
        """Function to change any or all of the parameters of the rectangle of
        the content.

        Args:
            x (int, float, optional): The x coordinate of the left of the
                rectangle.
            y (int, float, optional): The y coordinate of the top of the
                rectangle.
            width (int, float, optional): The width of the rectangle where the
                contents will be arranged.
            height (int, float, optional): The height of the rectangle where the
                contents will be arranged.
            last_part (int, optional): If not None, this is the index of the
                part that was being processed the last time method ``run`` was
                called.
            last_word (int, optional): If not None, this is the index of the
                word of the last part that was added the last time method
                ``run`` was called.
        """
        if x is not None:
            self.x = x
        if y is not None:
            self.y = y
        if width is not None:
            self.width = width
        if height is not None:
            self.height = height

    def init(self) -> None:
        """Function to reset all of the instance properties that have to be
        resetted before running the arranging process in a new rectangle.

        This function is called by method ``run``.
        """
        self.started = False
        self.lines = []
        self.text = ''
        self.graphics = ''
        self.ids = {}
        self.first_line_added = False

        self.used_fonts = set()
        self.current_line_used_fonts = set()
        self.current_height = 0
        self.lines = []

        line_width = self.width - (self.indent if self.is_first_line else 0)
        self.current_line = PDFTextLine(
            self.fonts, line_width, self.text_align, self.line_height
        )

        self.last_indent = 0
        self.last_state = self.last_factor = self.last_fill = None
        self.last_color = self.last_stroke_width = None

        self.y_ = 0

    def run(
        self, x: Number=None, y: Number=None, width: Number=None,
        height: Number=None
    ) -> dict:
        """Function to create the data needed to add this paragraph to the PDF
        document.

        This function will try to add all of the dict parts in ``content``
        argument list (or tuple) to this paragraph rectangle. Check this class
        documentation for more information about this method.

        This function args are the same as
        :meth:`pdfme.text.PDFTextBase.setup`.

        Returns:
            dict: The dict from the property ``result``.
        """
        self.setup(x, y, width, height)
        self.init()
        for part_index in range(self.last_part, len(self.content)):
            part = self.content[part_index]
            if not isinstance(part, dict):
                raise TypeError(
                    'elements in content must be of type dict: {}'
                    .format(part)
                )
            if 'type' in part:
                if part['type'] == 'br':
                    self.last_part_added = part_index + 1
                    self.last_word_added = 0
                    continue_ = self.add_current_line(True)
                    self.current_line = PDFTextLine(
                        self.fonts, self.width, self.text_align,
                        self.line_height
                    )
                    if not continue_:
                        return self.result
            else:
                continue_ = self.add_part(part, part_index)
                if not continue_:
                    return self.result

        continue_ = self.add_current_line(True)
        if continue_:
            self.finished = True
        return self.result

    def add_part(self, part: dict, part_index: int) -> bool:
        """Function used by methodm ``run`` to add one paragraph part at a time.

        Args:
            part (dict): part to be added.
            part_index (int): index of the part to be added.

        Returns:
            bool: whether it was able to add all of the parts (True) or the
                vertical space ran out.
        """

        if 'var' in part and self.pdf:
            part['text'] = str(self.pdf.context.get(part['var'], ''))

        words = part.get('text')
        if not isinstance(words, (str, list, tuple)):
            return 'continue'

        style = TEXT_DEFAULTS.copy()
        style.update(part.get('style', {}))
        new_line_part = self.current_line.add_line_part(
            style=style, ids=part.get('ids')
        )

        if not self.list_setup_done and self.list_text:
            self.list_setup_done = True
            self.setup_list()

        self.current_line_used_fonts.add((
            new_line_part.state.font_family,
            new_line_part.state.font_mode
        ))

        if isinstance(words, str):
            part['text'] = words = [
                ' ' if w.isspace() else w
                for w in re.split('( +)', words)
                if w != ''
            ]
        is_last_part = part_index == len(self.content) - 1

        for word_index in range(self.last_word, len(words)):
            word = words[word_index]
            ans = self.current_line.add_word(word)
            if ans['status'] == 'added':
                self.last_part_added = part_index
                self.last_word_added = word_index + 1
            elif ans['status'] == 'finished':
                continue_ = self.add_current_line(
                    is_last_part and word_index == len(words) - 1
                )
                ans['new_line'].max_width = self.width - (
                    self.list_indent if self.list_text else 0
                )
                ans['new_line'].next_line.max_width = ans['new_line'].max_width
                self.current_line = ans['new_line']
                if not continue_:
                    return False

        self.last_word = 0
        return True

    def add_current_line(self, is_last: bool=False) -> bool:
        """Function to add the current line to the list of already added lines.

        Args:
            is_last (bool, optional): whether this is the last line of this
                paragraph (True) or not (False).

        Returns:
            bool: whether this line was successfully added (True) or not (False)
        """
        if is_last and self.current_line.next_line is not None:
            self.current_line.line_parts.extend(
                self.current_line.next_line.line_parts
            )

        line_height = self.current_line.height
        if self.first_line_added:
            line_height *= self.line_height
        else:
            self.first_line_added = True

        if line_height + self.current_height > self.height:
            self.last_part = self.last_part_line
            self.last_word = self.last_word_line
            return False
        else:
            self.last_part_line = self.last_part_added
            self.last_word_line = self.last_word_added
            self.current_height += line_height
            self.lines.append(self.current_line)
            self.used_fonts.update(self.current_line_used_fonts)
            self.current_line_used_fonts = set()

            self.add_line_to_stream(self.current_line, is_last)
            self.current_line = None

            return True

    def setup_list(self) -> None:
        """This function is called when the first part of the paragraph is going
        to be added, and if this is a list paragraph, i.e. a paragraph with a
        something on its left (a bullet, a number, etc), this function will
        setup everything needed to display the text of the list paragraph, and
        will adjust its width to make space for the list text.

        Raises:
            TypeError: if list_style or list_indent passed to this instance
                are not a dict and an number respectively.
        """
        style = self.current_line.next_line.line_parts[0].style.copy()

        if self.list_style is None:
            self.list_style = {}
        elif isinstance(self.list_style, str):
            self.list_style = process_style(self.list_style, self.pdf)

        if not isinstance(self.list_style, dict):
            raise TypeError(
                'list_style must be a str or a dict. Value: {}'
                .format(self.list_style)
            )

        style.update(self.list_style)
        line_part = PDFTextLinePart(style, self.fonts)

        self.current_line_used_fonts.add(
            (line_part.state.font_family, line_part.state.font_mode)
        )

        if self.list_indent is None:
            self.list_indent = line_part.get_word_width(str(self.list_text))
        elif not isinstance(self.list_indent, (float, int)):
            raise TypeError(
                'list_indent must be int or float. Value: {}'
                .format(self.list_indent)
            )

        self.list_state = line_part.state
        self.current_line.max_width -= self.list_indent

    def add_line_to_stream(self, line: PDFTextLine, is_last:bool=False) -> None:
        """Function to add a PDFTextLine representing a paragraph line to the
        already added lines stream.

        Args:
            line (PDFTextLine): The line to be added to the stream.
            is_last (bool, optional): whether this is the last line of this
                paragraph (True) or not (False).
        """
        words_width, spaces_width = line.get_widths()
        x = self.list_indent if self.list_text else 0
        line_height = line.height
        full_line_height = line_height
        ignore_factor = self.text_align != 'j' or is_last or spaces_width == 0
        factor_width = self.width - words_width - x
        adjusted_indent = 0
        if self.text_align in ['r', 'c']:
            indent = self.width - words_width - spaces_width
            if self.text_align == 'c':
                indent /= 2
            x += indent
            adjusted_indent = indent - self.last_indent
            self.last_indent = indent

        if not self.started:
            self.started = True
            if self.is_first_line:
                factor_width -= self.indent
                x += self.indent
                self.is_first_line = False

                first_indent = (
                    adjusted_indent if self.text_align in ['r', 'c']
                    else self.indent
                )

                if self.list_text:
                    first_indent += self.list_indent
                    if self.list_state.size > full_line_height:
                        full_line_height = self.list_state.size
                    self.text += ' 0 -{} Td{} ({})Tj {} 0 Td'.format(
                        round(full_line_height, 3),
                        self.list_state.compare(self.last_state),
                        self.list_text, first_indent
                    )
                else:
                    self.text += ' {} -{} Td'.format(
                        round(first_indent, 3), round(full_line_height, 3)
                    )
            else:
                first_indent = adjusted_indent
                if self.list_text:
                    first_indent += self.list_indent
                self.text += ' {} -{} Td'.format(
                    round(first_indent, 3), round(full_line_height, 3)
                )
        else:
            if self.correct_indent:
                self.correct_indent = False
                adjusted_indent -= self.indent

            full_line_height *= self.line_height

            self.text += ' {} -{} Td'.format(round(adjusted_indent, 3),
                                        round(full_line_height, 3))

        self.y_ -= full_line_height

        factor = 1 if ignore_factor else factor_width / spaces_width

        for part in line.line_parts:
            text = self.clean_words(part.words)
            self.text += self.output_text(part, text, factor)
            part_width = part.current_width(factor)
            part_size = round(part.state.size, 3)

            if text != '' and not text.isspace():
                if part.ids is not None:
                    for id_ in part.ids:
                        id_y = self.y_ + part.state.rise - part_size*0.25
                        self.ids.setdefault(id_, []).append([
                            round(x, 3), round(id_y, 3),
                            round(x + part_width, 3), round(id_y + part_size, 3)
                        ])

                part_graphics = self.output_graphics(
                    part, x, self.y_, part_width
                )
                self.graphics += part_graphics
            x += part_width

    def clean_words(self, words: list) -> str:
        """This function joins a list of words (spaces included) and makes the
        resulting string compatible with a PDF string.

        Args:
            words (list): a list of strings, where each string is a word.

        Returns:
            str: A string with all of the words passed.
        """

        text = ''.join(word for word in words)
        if text != '':
            text = text.replace('\\',r'\\').replace('(','\(').replace(')','\)')
        return text

    def output_text(self, part: PDFTextLinePart, text, factor: Number=1) -> str:
        """Function that creates a piece of PDF stream (only the text), from
        the PDFTextLinePart and the ``text`` arguments.

        Args:
            part (PDFTextLinePart): the part to be transformed into a string
                representing a PDF stream piece.
            text ([type]): the text to be transformed into a string
                representing a PDF stream piece.
            factor (Number, optional): factor of the line needed to create
                center, right and justified aligned paragraphs.

        Returns:
            str: representing the PDF stream
        """
        stream = part.state.compare(self.last_state)
        self.last_state = part.state

        tw = round(part.space_width * (factor - 1), 3)
        if self.last_factor != tw:
            if tw == 0:
                tw = 0
            stream += ' {} Tw'.format(tw)
            self.last_factor = tw

        if text != '':
            # TODO: How can we add unicode to PDF string
            stream += ' ({})Tj'.format(text)
        return stream

    def output_graphics(
        self, part: PDFTextLinePart, x: Number, y: Number, part_width: Number
    ) -> str:
        """Function that creates a piece of PDF stream (only the graphics),
        from the PDFTextLinePart argument.

        Args:
            part (PDFTextLinePart): the part to be transformed into a string
                representing a PDF stream piece.
            x (int, float): the x origin coordinate of the graphics being added.
            y (int, float): the y origin coordinate of the graphics being added.
            width (int, float): the width of the part being added

        Returns:
            str: representing the PDF stream
        """
        graphics = ''
        if part.background is not None and not part.background.color is None:
            if part.background != self.last_fill:
                self.last_fill = part.background
                graphics += ' ' + str(self.last_fill)

            graphics += ' {} {} {} {} re F'.format(
                round(x, 3),
                round(y + part.state.rise - part.state.size*0.25, 3),
                round(part_width, 3), round(part.state.size*1.2, 3)
            )

        if part.underline:
            color = PDFColor(part.state.color, True)
            stroke_width = part.state.size * 0.07
            y_u = round(y + part.state.rise - stroke_width, 3)

            if color != self.last_color:
                self.last_color = color
                graphics += ' ' + str(self.last_color)

            if stroke_width != self.last_stroke_width:
                self.last_stroke_width = stroke_width
                graphics += ' {} w'.format(round(self.last_stroke_width, 3))

            graphics += ' {} {} m {} {} l S'.format(
                round(x, 3), y_u, round(x + part_width, 3), y_u
            )

        return graphics

class PDFText(PDFTextBase):
    """Class that represents a rich text paragraph to be added to a
    :class:`pdfme.pdf.PDF` instance.

    ``content`` argument should be a dict, with a key starting with a dot, like
    ``'.b;s:10;c:1;u'`` for example (keep reading to learn more about the format
    of this key), which we are going to refer to as the "the dot key" from here
    on. The value for the dot key is a list/tuple containing strings or more
    ``content`` dicts like the one we are describing here (you can have nested
    ``content`` dicts or what we call a paragraph part), but for simplicity,
    you can pass a string (for non-rich text) or a tuple/list with strings and
    more paragraph parts:

    * If ``content`` argument is a string, it will become the following:

      .. code-block:: python

          { '.': [ <STRING>, ] }

    * If ``content`` argument is a list/tuple, it will become the following:

      .. code-block:: python

          { '.': <LIST_OR_TUPLE> }

    This is an example of a ``content`` argument:

    .. code-block:: python

        {
            ".b;u;i;c:1;bg:0.5;f:Courier": [
                "First part of the paragraph ",
                {
                    ".b:0;u:0;i:0;c:0;bg:": [
                        "and here the second, nested inside the root paragraph,",
                    ]
                },
                "and yet one more part before a ",
                {".c:blue;u:1": "url link", "uri": "https://some.url.com"}
            ]
        }

    This class is a subclass of :class:`pdfme.text.PDFTextBase` and adds the
    logic to let the user of this class pass content in a nested cascading
    "jsonish" format (like HTML), i.e. if you pass a dict to ``content``,
    and this dict has a ``style`` key, all of its children will inherit this
    style and will be able to overwrite some or all of the style parameters
    coming from it. The children will be able to pass their own
    style parameters to their children too, and so on.

    Additional to the dot key, paragraph parts can have the following keys:

    * ``'label'``: this is a string with a unique name (there should be
      only one label with this name in the whole document) representing
      a destination that can be referenced in other parts of the
      document. This is suited for titles, figures, tables, etc.

    * ``'ref'``: this is a string with the name of a label, that will
      become a link to the position of the label referenced.

    * ``'uri'``: this is a string with a reference to a web resource,
      that will turn this text part in a link to that web page.

    * ``'outline'``: an outline is a label that is shown in the outlines panel
      of the PDF reader. This outlines show the structure of the document.
      This attribute is a dict with the following optional keys:
      
      * ``level``: an optional int with the level of this outline in the
        outlines tree. The default value is 1.
        
      * ``text``: an optional string to be shown in the outlines panel for this
        outline. The default value is the contents of this part.

    * ``'ids'``: when method ``run`` is called, dict attr ``result`` is
      available with information to add the paragraph to the PDF, and
      within that information you'll find a key ``ids``, a dict with
      the position and size of the rectangle for each of the ids you
      include in this argument. This way you can "tag" a part of a
      paragraph, call ``run``, and get the position of it afterwards.

    * ``'var'``: this is a string with the name of a global variable,
      previously set in the containing :class:`pdfme.pdf.PDF`
      instance, by adding a new key to its dict attribute ``context``.
      This way you can reuse a repetitive string throughout the PDF
      document.

    Style of the paragraph dicts can be defined in the dot key
    itself (a string with a semi-colon separeted list of the attributes,
    explained in :func:`pdfme.utils.parse_style_str`) or in a ``style`` dict
    too. The attributes for a paragraph style are the following:

    * ``'b'`` (bool) to make text inside this part bold. Default is False.

    * ``'i'`` (bool) to make text inside this part cursive (italics,
      oblique). Default is False.

    * ``'s'`` (int, float) to set the size of the text inside this
      part. Default is 11.

    * ``'f'`` (str) to set the font family of the text inside this
      part. Default is ``'Helvetica'``.

    * ``'u'`` (bool) to make the text inside this part underlined.
      Default is False.

    * ``'c'`` (int, float, list, tuple, str) to set the color of
      the text inside this part. See :func:`pdfme.color.parse_color`
      for information about this attribute. Default is black.

    * ``'bg'`` (int, float, list, tuple, str) to set the background
      color of the text inside this part. See
      :func:`pdfme.color.parse_color` for information about this
      attribute. Default is None.

    * ``'r'`` (int, float) to set the baseline of the text, relative
      to the normal baseline. This is a fraction of the current size of
      the text, i.e. it will move the baseline the text size times this
      number in points, upwards if positive, and downwards if negative.
      Default is 0.

    One more example of a ``content`` argument with a ``style`` dict, and
    additional keys:

    .. code-block:: python

        {
            '.': ['text to be displayed'],
            'style': {
                'b': True,
                'i': True,
                'u': True,
                's': 10.2,
                'f': 'Courier',
                'c': 0.9,
                'bg': 'red',
                'r': 0.5
            },
            'label': 'a_important_paragraph',
            'uri': 'https://github.com/aFelipeSP/pdfme'
        }

    With arguments ``list_text``, ``list_indent`` and ``list_style`` you can
    turn a paragraph into a list paragraph, one that has a bullet or a number
    at the left of the paragraph, with an additional indentation. With this
    you can build a bulleted or numbered list of paragraphs.

    Args:
        content (str, list, tuple, dict): the one just described.
        width (int, float): The width of the paragraph.
        height (int, float): The height of the paragraph.
        x (int, float, optional): The x coordinate of the paragraph.
        y (int, float, optional): The y coordinate of the paragraph
        fonts (PDFFonts, optional): To extract information about the fonts
            being used in the paragraph.
        text_align (str, optional): ``'l'`` for left (default), ``'c'`` for
            center, ``'r'`` for right and ``'j'`` for justified text.
        line_height (int, float, optional): How much space between the
            lines of the paragraph. This is a fraction of each line's
            height, so the real distance between lines can vary depending on
            the text size of each part of the paragraph.
        indent (int, float, optional): The indentation of the first line of
            the paragraph.
        list_text (str, optional): Needed if you want to turn this paragraph
            into a list paragraph. This text will be displayed before the
            paragraph and will be aligned with the first line.
        list_indent (int, float, optional): Needed if you want to turn this
            paragraph into a list paragraph. The space between the start of
            the left side of the rectangle and the left side of the
            paragraph itself. If omitted, this space will be the width of
            the ``list_text``.
        list_style (dict, optional): Needed if you want to turn this
            paragraph into a list paragraph. The style of ``list_text``.
            If omitted, the style of the first part of the first line will
            be used.
        pdf (PDF, optional): To grab global information of the PDF being
            used.
    """
    def __init__(
        self, content: ContentType, width: Number, height: Number,
        x: Number=0, y: Number=0, fonts: 'PDFFonts'=None, text_align: str=None,
        line_height: Number=None, indent: Number=0, list_text: str=None,
        list_indent: Number=None, list_style: dict=None, pdf: 'PDF'=None
    ) -> None:
        self.pdf = pdf
        self.fonts = fonts
        self.content = []
        self._recursive_content_parse(content, TEXT_DEFAULTS, [])
        super().__init__(
            self.content, width, height, x, y, fonts, text_align, line_height,
            indent, list_text, list_indent, list_style, pdf
        )

    def _new_text_part(
        self, style: dict, ids: list, part_var: str=None, last_part: dict=None
    ) -> dict:
        """Creates a new text part compatible with
        :class:`pdfme.text.PDFTextBase`.

        Args:
            style (dict): The style of this new part.
            ids (list): The ids of this new part.
            part_var (str, optional): The name of the 'var' (None if there's no
                var for this part).
            last_part (dict, optional): the part before the one that is going to
                be created with this function.

        Returns:
            dict: representing a part compatible with
                :class:`pdfme.text.PDFTextBase`.
        """
        if last_part is not None and last_part['text'] == '':
            self.content.remove(last_part)
        text_part = {'style': style, 'text': '', 'ids': ids}
        if part_var is not None:
            text_part['var'] = part_var
        self.content.append(text_part)
        return text_part

    def _recursive_content_parse(
        self, content: ContentType, parent_style: dict, ids: list
    ) -> None:
        """Function to be called recursively by this class, to transform the
        content passed to this instance into a list of parts compatible with
        :class:`pdfme.text.PDFTextBase`.

        Args:
            content (str, list, tuple, dict): An object like the one explained
                in the documentation of this class.
            parent_style (dict): A dict with the style of the parent of the
                current part being parsed.
            ids (list): A list of the ids of the parent of the current part
                being parsed.

        Raises:
            TypeError: If the content part passed doesn't have the format
                described in the documentation of this class.
        """
        style = parent_style.copy()
        ids = ids.copy()

        if isinstance(content, str):
            content = {'.': [content]}
        elif isinstance(content, (list, tuple)):
            content = {'.': content}

        if not isinstance(content, dict):
            raise TypeError(
                'content must be of type dict, str, list or tuple: {}'
                .format(content)
            )

        elements = []
        for key, value in content.items():
            if key.startswith('.'):
                style.update(parse_style_str(key[1:], self.fonts))
                if isinstance(value, (int, float)):
                    value = [str(value)]
                elif isinstance(value, str):
                    value = [value]
                if not isinstance(value, (list, tuple)):
                    raise TypeError(
                        'value of . attr must be of type str, list or tuple: {}'
                        .format(value)
                    )
                elements = value
                break

        style.update(process_style(content.get('style'), self.pdf))
        part_var = content.get('var')
        text_part = self._new_text_part(style, ids, part_var)
        text_part['ids'].extend(content.get('ids', []))

        if part_var is not None:
            elements = ['0']

        label = content.get('label')
        if label is not None:
            text_part['ids'].append('$label:' + label)
        ref = content.get('ref')
        if ref is not None:
            text_part['ids'].append('$ref:' + ref)
        uri = content.get('uri')
        if uri is not None:
            text_part['ids'].append('$uri:' + uri)
        outline = content.get('outline')
        if isinstance(outline, dict):
            text = outline.get('text', ''.join(str(e) for e in elements))
            level = outline.get('level', 1)
            if label is None:
                outline_label = str(uuid4())
                text_part['ids'].append('$label:' + outline_label)
            else:
                outline_label = label
            outline_ = {'text': text, 'level': level, 'label': outline_label}
            text_part['ids'].append('$outline:{}'.format(json.dumps(outline_)))

        is_last_string = False

        for element in elements:
            if isinstance(element, (int, float)):
                element = str(element)
            if isinstance(element, str):
                if element == '':
                    continue
                lines = element.split('\n')
                if not is_last_string:
                    text_part = self._new_text_part(
                        style, text_part['ids'], part_var, text_part
                    )
                text_part['text'] += lines[0]
                for line in lines[1:]:
                    self.content.append({'type': 'br'})
                    text_part = self._new_text_part(
                        style, text_part['ids'], part_var, text_part
                    )
                    text_part['text'] += line
                is_last_string = True
            elif isinstance(element, dict):
                self._recursive_content_parse(element, style, text_part['ids'])
                is_last_string = False
            else:
                raise TypeError(
                    'elements must be of type str or dict: {}'.format(element)
                )

        if text_part is not None and text_part['text'] == '':
            self.content.remove(text_part)

from .color import PDFColor
from .fonts import PDFFonts
from .pdf import PDF
from .utils import get_paragraph_stream, parse_style_str, process_style, copy



import json
from io import BytesIO
from pathlib import Path
from typing import Any, Iterable, Union

Number = Union[int, float]
PageType = Union[Number, str, Iterable[Number]]
MarginType = Union[int, float, Iterable[Number], dict]
ColorType = Union[int, float, str, list, tuple]
ImageType = Union[str, Path, BytesIO]
TextType = Union[str, list, tuple, dict]
class PDF:
    """Class that represents a PDF document, and has methods to add pages,
    and to add paragraphs, images, tables and a mix of this, a content box,
    to them.

    You can use this class to create a PDF file, by adding one page at a
    time, and adding stuff to each page you add, like this:

    .. code-block:: python

        from pdfme import PDF

        pdf = PDF()
        pdf.add_page()
        pdf.text('This is a paragraph')

        with open('document.pdf', 'wb') as f:
            pdf.output(f)

    Through the constructor arguments you can modify the default features
    of the PDF document, like the size of the pages, their orientation,
    the page numbering options, and the appearance of the text. These are
    used everytime you create a new page, or a new paragraph, but you
    can overwrite these for each case.

    You can change the default values for the pages by calling
    :meth:`pdfme.pdf.PDF.setup_page`, and change the default values for text
    by changing attributes ``font_family``, ``font_size``, ``font_color``,
    ``text_align`` and ``line_height``.

    Methods :meth:`pdfme.pdf.PDF.text`, :meth:`pdfme.pdf.PDF.image`,
    :meth:`pdfme.pdf.PDF.table` and :meth:`pdfme.pdf.PDF.content` are the main
    functions to add paragraphs, images, tables and content boxes respectively,
    and all of them, except the image method, take into account the margins of
    the current page you are working on, and create new pages automatically if
    the stuff you are adding needs more than one page. If you want to be
    specific about the position and the size of the paragraphs, tables and
    content boxes you are inserting, you can use methods
    :meth:`pdfme.pdf.PDF._text`, :meth:`pdfme.pdf.PDF._table` and
    :meth:`pdfme.pdf.PDF._content` instead, but these don't handle the creation
    of new pages like the first ones.

    Each page has attributes ``x`` and ``y`` that are used to place elements
    inside them, and for the methods that receive ``x`` and ``y`` arguments,
    if they are None, the page's ``x`` and ``y`` attributes are used instead.

    For more information about paragraphs see :class:`pdfme.text.PDFText`, and
    about tables :class:`pdfme.table.PDFTable`.

    Although you can add all of the elements explained so far, we recommend
    using content boxes only, because all of the additional funcionalities they
    have, including its ability to embed other elements. For more information
    about content boxes see :class:`pdfme.content.PDFContent`.

    Paragraphs, tables and content boxes use styles to give format to the
    content inside of them, and sometimes styling can get repetitive. This is
    why there's a dict attribute called ``formats`` where you can add named
    style dicts and used them everywhere inside this document, like this:


    .. code-block:: python

        from pdfme import PDF

        pdf = PDF()
        pdf.formats['link'] = {
            'c': 'blue',
            'u': True
        }
        pdf.add_page()
        pdf.text({
            '.': 'this is a link',
            'style': 'link',
            'uri': 'https://some.domain.com'
        })

    If you find yourself using a piece of text often in the document, you can
    add it to the dict attribute ``context`` and include it in any paragraph in
    the document by using its key in the dict, like this:

    .. code-block:: python

        from pdfme import PDF

        pdf = PDF()
        pdf.context['arln'] = 'A Really Long Name'
        pdf.add_page()
        pdf.text({
            '.': ['The following name is ', {'var': 'arln'}, '.']
        })

    There are some special ``context`` variables that are used by us that start
    with symbol ``$``, so it's adviced to name your own variables without this
    symbol in the beginning. The only of these variables you should care about
    is ``$page`` that contains the number of the current page.

    You can add as much running sections as you want by using
    :meth:`pdfme.pdf.PDF.add_running_section`. Running sections are
    content boxes that are included on every page you create after adding them.
    Through these you can add a header and a footer to the PDF.

    If you want a simpler and more powerful interface, you should use
    :class:`pdfme.document.PDFDocument`.

    Args:
        page_size (str, int, float, tuple, list, optional): this argument sets
            the dimensions of the page. See :func:`pdfme.utils.get_page_size`.
        rotate_page (bool, optional): whether the page dimensions should be
            inverted (True), or not (False).
        margin (str, int, float, tuple, list, dict, optional): the margins of
            the pages. See :func:`pdfme.utils.parse_margin`.
        page_numbering_offset (int, float, optional): if the number of the page
            is included, this argument will set the offset of the page. For
            example if the current page is the 4th one, and the offset is 3, the
            page number displayed in the current page will be 1.
        page_numbering_style (str, optional): the style of the page number.
            Options are ``arabic`` (1,2,3,...) and ``roman`` (I, II, III, IV,
            ...).
        font_family (str, optional): The name of the font family. Options are
            ``Helvetica`` (default), ``Times``, ``Courier``, ``Symbol`` and
            ``ZapfDingbats``. You will also be able to add new fonts in a future
            release.
        font_size (in, optional): The size of the font.
        font_color (int, float, str, list, tuple, optional): The color of the
            font. See :func:`pdfme.color.parse_color`.
        text_align (str, optional):  ``'l'`` for left (default), ``'c'`` for
            center, ``'r'`` for right and ``'j'`` for justified text.
        line_height (int, float, optional): space between the lines of the
            paragraph. See :class:`pdfme.text.PDFText`.
        indent (int, float, optional): space between left of the paragraph, and
            the beggining of the first line. See :class:`pdfme.text.PDFText`.
        outlines_level (int, optional): the level of the outlines to be
            displayed on the outlines panel when the PDF document is opened.
    """
    def __init__(
        self, page_size: PageType='a4', rotate_page: bool=False,
        margin: MarginType=56.693, page_numbering_offset: Number=0,
        page_numbering_style: str='arabic', font_family: str='Helvetica',
        font_size: Number=11, font_color: ColorType=0.1, text_align: str='l',
        line_height: Number=1.1, indent: Number=0, outlines_level: int=1
    ) -> None:
        self.setup_page(page_size, rotate_page, margin)
        self.page_numbering_offset = page_numbering_offset
        self.page_numbering_style = page_numbering_style

        self.font_family = font_family
        self.font_size = font_size
        self.font_color = font_color
        self.text_align = text_align
        self.line_height = line_height
        self.indent = indent
        self.outlines_level = outlines_level

        self.formats = {}
        self.context = {}

        self.dests = {}
        self.uris = {}
        self.pages = []
        self.running_sections = []
        self.outlines = []

        self.base = PDFBase()
        self.root = self.base.add({ 'Type': b'/Catalog'})
        self.base.trailer['Root'] = self.root.id

        self.fonts = PDFFonts()
        self.used_fonts = {}
        self.images = {}
        self._add_or_get_font('Helvetica', 'n')

    @property
    def page(self) -> 'PDFPage':
        """
        Returns:
            PDFPage: current page
        """
        return self.pages[self._page_index]

    @property
    def page_index(self) -> int:
        """
        Returns:
            int: current page index.
        """
        return self._page_index

    @page_index.setter
    def page_index(self, page_index):
        self._page_index = page_index
        self.context['$page'] = self.get_page_number()

    @property
    def width(self) -> float:
        """
        Returns:
            float: current page width
        """
        return self.page.width

    @property
    def height(self) -> float:
        """
        Returns:
            float: current page height
        """
        return self.page.height

    def setup_page(
        self, page_size: PageType=None, rotate_page: bool=None,
        margin: MarginType=None
    ) -> None:
        """Method to set the page features defaults. These values will be used
        from now on when adding new pages.

        Args:
            page_size (str, int, float, tuple, list, optional): this argument
                sets the dimensions of the page.
                See :func:`pdfme.utils.get_page_size`.
            rotate_page (bool, optional): whether the page dimensions should be
                inverted (True), or not (False).
            margin (str, int, float, tuple, list, dict, optional): the margins
                of the pages. See :func:`pdfme.utils.parse_margin`.
        """
        if page_size is not None:
            self.page_width, self.page_height = get_page_size(page_size)
        if rotate_page is not None:
            self.rotate_page = rotate_page
        if margin is not None:
            self.margin = parse_margin(margin)

    def add_page(
        self, page_size: PageType=None, rotate_page: bool=None,
        margin: MarginType=None
    ) -> None:
        """Method to add a new page. If provided, arguments will only apply for
        the page being added.

        Args:
            page_size (str, int, float, tuple, list, optional): this argument
                sets the dimensions of the page.
                See :func:`pdfme.utils.get_page_size`.
            rotate_page (bool, optional): whether the page dimensions should be
                inverted (True), or not (False).
            margin (str, int, float, tuple, list, dict, optional): the margins
                of the page. See :func:`pdfme.utils.parse_margin`.
        """
        if page_size is not None:
            page_width, page_height = get_page_size(page_size)
        else:
            page_height, page_width = self.page_height, self.page_width

        if (rotate_page is None and self.rotate_page) or rotate_page:
            page_height, page_width = page_width, page_height

        margin_ = copy(self.margin)
        if margin is not None:
            margin_.update(parse_margin(margin))

        page = PDFPage(self.base, page_width, page_height,
            **{'margin_' + side: value for side, value in margin_.items()}
        )

        self.pages.append(page)
        self.page_index = len(self.pages) - 1

        for running_section in self.running_sections:
            self._content(**running_section)

        page.go_to_beginning()

    def add_running_section(
        self, content: dict, width: Number, height: Number, x: Number, y:Number
    ) -> None:
        """Method to add running sections, like a header and a footer, to this
        document.

        Running sections are content boxes that are included on every page you
        create after adding them.

        Args:
            content (dict): a content dict like the one you pass to create a
                instance of :class:`pdfme.content.PDFContent`.
            width (int, float, optional): The width of the rectangle where the
                contents will be arranged.
            height (int, float, optional): The height of the rectangle where the
                contents will be arranged.
            x (int, float, optional): The x position of the left of the
                rectangle.
            y (int, float, optional): The y position of the top of the
                rectangle.
        """
        self.running_sections.append(dict(
            content=content, width=width, height=height, x=x, y=y
        ))

    def add_font(
        self, fontfile: str, font_family: str, mode: str='n'
    ) -> None:
        """Method to add a new font to this document. This functionality is not
        ready yet.

        Args:
            fontfile (str): the path of the fontfile.
            font_family (str): the name of the font family being added.
            mode (str, optional): the mode of the font being added. It can be
                ``n`` for normal, ``b`` for bold and ``i`` for italics
                (oblique).
        """
        self.fonts.load_font(fontfile, font_family, mode)

    def _add_or_get_font(self, font_family: str, mode: str) -> tuple:
        """Method to add a new font to the already used fonts list, if it has
        not been added, that returns information about the font to be used in
        other places in the PDF.

        Args:
            font_family (str): the name of the font family being used.
            mode (str, optional): the mode of the font being used. It can be
                ``n`` for normal, ``b`` for bold and ``i`` for italics
                (oblique).

        Returns:
            tuple: tuple with first element as the name of the font (a name used
            in the PDF text streams), and the second as the id of the inner
            PDF object that represents the font.
        """
        f = (font_family, mode)
        if f in self.used_fonts:
            return self.used_fonts[f]
        font = self.fonts.get_font(*f)
        font_obj = font.add_font(self.base)
        self.used_fonts[(font_family, mode)] = (font.ref, font_obj.id)
        return font.ref, font_obj.id

    def _used_font(self, font_family: str, mode: str) -> None:
        """Method to add a font to the current page.

        Args:
            font_family (str): the name of the font family being used.
            mode (str, optional): the mode of the font being used. It can be
                ``n`` for normal, ``b`` for bold and ``i`` for italics
                (oblique).
        """
        f = (font_family, mode)
        font_args = self._add_or_get_font(*f)
        self.page.add_font(*font_args)

    def create_image(
        self, image: ImageType, extension: str=None, image_name: str=None
    ) -> 'PDFImage':
        """Method to create a PDF image.

        Arguments for this method are the same as :class:`pdfme.image.PDFImage`.

        Returns:
            PDFImage: object representing the PDF image.
        """
        return PDFImage(image, extension, image_name)

    def add_image(
        self, pdf_image: 'PDFImage', x: Number=None, y: Number=None,
        width: Number=None, height:Number=None, move: str='bottom'
    ) -> None:
        """Method to add a PDF image to the current page.

        Args:
            pdf_image (PDFImage): the PDF image.
            x (int, float, optional): The x position of the left of the
                image.
            y (int, float, optional): The y position of the top of the
                image.
            width (int, float, optional): The width of the image. If this and
                ``height`` are None, the width will be the same as the page
                content width, but if this is None and ``height`` is not, the
                width will be calculated from ``height``, keeping the proportion
                of the image.
            height (int, float, optional): The height of the image. If this is
                None, the height will be calculated from the image ``width``,
                keeping the proportion.
            move (str, optional): wheter it should move page x coordinate to
                the right side of the image (``next``) or if it should move
                page y coordinate to the bottom of the image (``bottom``)
                (default).
        """
        if pdf_image.image_name not in self.images:
            image_obj = self.base.add(pdf_image.pdf_obj)
            self.images[pdf_image.image_name] = image_obj.id
        else:
            image_obj = self.base[self.images[pdf_image.image_name]]

        h = pdf_image.height
        w = pdf_image.width

        if width is None and height is None:
            width = self.page.content_width
            height = width * h/w
        elif width is None:
            width = height * w/h
        elif height is None:
            height = width * h/w

        if x is not None:
            self.page.x = x
        if y is not None:
            self.page._y = y

        self.page.add_image(image_obj.id, width, height)

        if move == 'bottom':
            self.page.y += height
        if move == 'next':
            self.page.x += width

    def image(
        self, image: ImageType, extension: str=None, image_name: str=None,
        x: Number=None, y: Number=None, width: Number=None, height:Number=None,
        move: str='bottom'
    ) -> None:
        """Method to create and add a PDF image to the current page.

        Args:
            image (str, Path, BytesIO): see :class:`pdfme.image.PDFImage`.
            extension (str, optional): see :class:`pdfme.image.PDFImage`.
            image_name (str, optional): see :class:`pdfme.image.PDFImage`.
            x (int, float, optional): the x position of the left of the
                image.
            y (int, float, optional): the y position of the top of the
                image.
            width (int, float, optional): the width of the image. If this and
                ``height`` are None, the width will be the same as the page
                content width, but if this is None and ``height`` is not, the
                width will be calculated from ``height``, keeping the proportion
                of the image.
            height (int, float, optional): the height of the image. If this is
                None, the height will be calculated from the image ``width``,
                keeping the proportion.
            move (str, optional): wheter it should move page x coordinate to
                the right side of the image (``next``) or if it should move
                page y coordinate to the bottom of the image (``bottom``)
                (default).
        """
        pdf_image = self.create_image(image, extension, image_name)
        self.add_image(
            pdf_image, x=x, y=y, width=width, height=height, move=move
        )

    def _default_paragraph_style(
        self, width: Number=None, height:Number=None, text_align: str=None,
        line_height: Number=None, indent: Number=None
    ) -> dict:
        """This method returns a dict with each of the arguments as the keys.
        If they are None, the PDF default value for each of them is used.

        For more information about the arguments of this method see
        :class:`pdfme.text.PDFText`

        Args:
            width (Number, optional): the width of the paragraph.
            height (Number, optional): the height of the paragraph.
            text_align (str, optional): the text align of the paragraph.
            line_height (Number, optional): the line height of the paragraph.
            indent (Number, optional): the indent of the paragraph.

        Returns:
            dict: dict with the default and the given parameters combined.
        """
        return dict(
            width = self.page.width - self.page.margin_right - self.page.x \
                if width is None else width,
            height = self.page.height - self.page.margin_bottom - self.page.y \
                if height is None else height,
            text_align = self.text_align if text_align is None else text_align,
            line_height = self.line_height if line_height is None \
                else line_height,
            indent = self.indent if indent is None else indent,
        )

    def _init_text(self, content: TextType) -> dict:
        """Method that prepares the paragraph passed as argument to be used in
        other methods.

        Args:
            content (str, list, tuple, dict): the paragraph object.

        Returns:
            dict: dict with the prepared paragraph.
        """
        style = {
            'f': self.font_family, 's': self.font_size, 'c': self.font_color
        }
        if isinstance(content, str):
            content = {'style': style, '.': [content]}
        elif isinstance(content, (list, tuple)):
            content = {'style': style, '.': content}
        elif isinstance(content, dict):
            style_str = [k[1:] for k in content.keys() if k.startswith('.')]
            if len(style_str) > 0:
                style.update(parse_style_str(style_str[0], self.fonts))
            style.update(process_style(content.get('style'), self))
            content['style'] = style
        return content

    def _position_and_size(
        self, x: Number=None, y: Number=None, width: Number=None,
        height: Number=None
    ) -> tuple:
        """Method that returns a tuple with the arguments passed, with default
        values if they are None.

        Args:
            x (int, float, optional): The x position of the left of the
                element.
            y (int, float, optional): The y position of the top of the
                element.
            width (int, float, optional): The width of the element.
            height (int, float, optional): The height of the element.

        Returns:
            tuple: tuple with the new ``x``,  ``y``, ``width`` and ``height``.
        """
        if x is not None:
            self.page.x = x
        if y is not None:
            self.page.y = y
        if width is None:
            width = self.page.width - self.page.margin_right - self.page.x
        if height is None:
            height = self.page.height - self.page.margin_bottom - self.page.y
        return self.page.x, self.page._y, width, height

    def get_page_number(self) -> str:
        """Method that returns the string reprensentation of the number of the
        current page.

        Returns:
            str: string with the page number that depends on attributes
            ``page_numbering_offset`` and ``page_numbering_style``.
        """
        page = self.page_index + 1 + self.page_numbering_offset
        return to_roman(page) if self.page_numbering_style == 'roman'\
            else str(page)

    def _create_text(
        self, content: TextType, width: Number=None, height: Number=None,
        text_align: str=None, line_height: Number=1.1, indent: Number=0,
        list_text: str=None, list_indent: Number=None, list_style: dict=None
    ) -> 'PDFText':
        """Method to create a paragraph.

        For more information about the arguments see
        :class:`pdfme.text.PDFText`.

        Returns:
            PDFText: object that represents a paragraph.
        """
        par_style = self._default_paragraph_style(
            width, height, text_align, line_height, indent
        )
        par_style.update({
            'list_text': list_text, 'list_indent': list_indent,
            'list_style': list_style
        })
        content = self._init_text(content)
        pdf_text = PDFText(content, fonts=self.fonts, pdf=self, **par_style)
        return pdf_text

    def _add_text(
        self, text_stream: str, x: Number=None, y: Number=None,
        width: Number=None, height: Number=None, graphics_stream: str=None,
        used_fonts: tuple=None, ids: dict=None, move: str='bottom'
    ) -> None:
        """Method to add a paragraph to the current page. The arguments for this
        method are obtained from the attribute ``result`` of class
        :class:`pdfme.text.PDFText`.

        For information about the arguments of this method see
        :meth:`pdfme.text.PDFTextBase.result`
        """
        stream = get_paragraph_stream(x, y, text_stream, graphics_stream)
        self.page.add(stream)

        for font in used_fonts:
            self._used_font(*font)

        for id_, rects in ids.items():
            if len(rects) == 0:
                continue
            if id_.startswith('$label:'):
                d = rects[0]
                x_ref = x + d[0]
                self.dests[id_[7:]] = [
                    self.page.page.id, b'/XYZ', round(x_ref, 3),
                    round(y + d[3], 3), round(x_ref/self.page.width, 3) + 1
                ]
            elif id_.startswith('$ref:'):
                for r in rects:
                    self.page.add_reference(
                        id_[5:],
                        [
                            round(x + r[0], 3), round(y + r[1], 3),
                            round(x + r[2], 3), round(y + r[3], 3)
                        ]
                    )
            elif id_.startswith('$uri:'):
                link = id_[5:]
                if not link in self.uris:
                    uri = self.base.add(
                        {'Type': b'/Action', 'S': b'/URI', 'URI': link}
                    )
                    self.uris[link] = uri.id

                for r in rects:
                    self.page.add_link(
                        self.uris[link],
                        [
                            round(x + r[0], 3), round(y + r[1], 3),
                            round(x + r[2], 3), round(y + r[3], 3)
                        ]
                    )
            elif id_.startswith('$outline:'):
                outline_data = json.loads(id_[9:])
                outline = self.outlines
                for _ in range(outline_data['level'] - 1):
                    outline = outline[-1].setdefault('children', [])

                outline.append(outline_data)

        if move == 'bottom':
            self.page.y += height
        if move == 'next':
            self.page.x += width

    def _text(
        self, content: Union[TextType, 'PDFText'], width: Number=None,
        height: Number=None, x: Number=None, y: Number=None,
        text_align: str=None, line_height: Number=1.1, indent: Number=0,
        list_text: str=None, list_indent: Number=None, list_style: dict=None,
        move: str='bottom'
    ) -> 'PDFText':
        """Method to create and add a paragraph to the current page.

        If ``content`` is a PDFText, the method ``run`` for this instance will
        be called with the new rectangle passed to this function.
        Else, this method will try to build a new PDFText instance with argument
        ``content`` and call method ``run`` afterwards.

        For more information about the arguments see
        :class:`pdfme.text.PDFText`.

        Returns:
            PDFText: object that represents the paragraph.
        """
        x, y, width, height = self._position_and_size(x, y, width, height)
        if isinstance(content, PDFText):
            pdf_text = content
            pdf_text.run(x, y, width, height)
        else:
            pdf_text = self._create_text(
                content, width, height, text_align, line_height, indent,
                list_text, list_indent, list_style
            )
            pdf_text.run(x, y)

        self._add_text(move=move, **pdf_text.result)
        return pdf_text

    def text(
        self, content: TextType, text_align: str=None, line_height: Number=1.1,
        indent: Number=0, list_text: str=None, list_indent: Number=None,
        list_style: dict=None
    ) -> None:
        """Method to create and add a paragraph to this document. This method
        will keep adding pages to the PDF until all the contents of the
        paragraph are added to the document.

        For more information about the arguments see
        :class:`pdfme.text.PDFText`.
        """
        pdf_text = self._text(
            content, x=self.page.margin_left, width=self.page.content_width,
            text_align=text_align, line_height=line_height, indent=indent,
            list_text=list_text, list_indent=list_indent, list_style=list_style
        )
        while not pdf_text.finished:
            self.add_page()
            pdf_text = self._text(pdf_text, self.page.content_width,
                self.page.content_height, self.page.margin_left,
                self.page.margin_top
            )

    def _default_content_style(self) -> dict:
        """This method returns a dict with the PDF default values used by
        content boxes.

        For more information about the keys in the dict returned by this method
        see :class:`pdfme.text.PDFText`

        Returns:
            dict: dict with the default values.
        """
        return dict(
            f=self.font_family, s=self.font_size, c=self.font_color,
            text_align=self.text_align, line_height=self.line_height,
            indent=self.indent
        )

    def _create_table(
        self, content: Iterable, width: Number=None, height: Number=None,
        x: Number=None, y: Number=None, widths: Iterable=None,
        style: Union[dict, str]=None, borders: Iterable=None,
        fills: Iterable=None
    ) -> 'PDFTable':
        """Method to create a table.

        For more information about this method arguments see
        :class:`pdfme.table.PDFTable`.

        Returns:
            PDFTable: object that represents a table.
        """
        style_ = self._default_content_style()
        style_.update(process_style(style, self))
        pdf_table = PDFTable(
            content, self.fonts, x, y, width, height,
            widths, style_, borders, fills, self
        )
        return pdf_table

    def _table(
        self, content: Union[Iterable, 'PDFTable'], width: Number=None,
        height: Number=None, x: Number=None, y: Number=None,
        widths: Iterable=None, style: Union[dict, str]=None,
        borders: Iterable=None, fills: Iterable=None, move: str='bottom'
    ) -> 'PDFTable':
        """Method to create and add a table to the current page.

        If ``content`` is a PDFTable, the method ``run`` for this instance will
        be called with the new rectangle passed to this function.
        Else, this method will try to build a new PDFTable instance with
        argument ``content`` and call method ``run`` afterwards.

        For more information about this method arguments see
        :class:`pdfme.table.PDFTable`.

        Returns:
            PDFTable: object that represents a table.
        """

        x, y, width, height = self._position_and_size(x, y, width, height)

        if isinstance(content, PDFTable):
            pdf_table = content
            pdf_table.run(x, y, width, height)
        else:
            pdf_table = self._create_table(
                content, width, height, x, y, widths, style, borders, fills
            )
            pdf_table.run()

        self._add_graphics([*pdf_table.fills, *pdf_table.lines])
        self._add_parts(pdf_table.parts)

        if move == 'bottom':
            self.page.y += pdf_table.current_height
        if move == 'next':
            self.page.x += width

        return pdf_table

    def table(
        self, content: Iterable, widths: Iterable=None,
        style: Union[str, dict]=None, borders: Iterable=None,
        fills: Iterable=None
    ) -> None:
        """Method to create and add a table to this document. This method
        will keep adding pages to the PDF until all the contents of the
        table are added to the document.

        For more information about this method arguments see
        :class:`pdfme.table.PDFTable`.
        """
        pdf_table = self._table(
            content, widths=widths, style=style, borders=borders, fills=fills,
            x=self.page.margin_left, width=self.page.content_width
        )
        while not pdf_table.finished:
            self.add_page()
            pdf_table = self._table(
                pdf_table, self.page.content_width, self.page.content_height,
                self.page.margin_left, self.page.margin_top
            )

    def _create_content(
        self, content: dict, width: Number=None, height: Number=None,
        x: Number=None, y: Number=None
    ) -> 'PDFContent':
        """Method to create a content box.

        For more information about this method arguments see
        :class:`pdfme.content.PDFContent`.

        Returns:
            PDFContent: object that represents a content box.
        """
        style = self._default_content_style()
        content = content.copy()
        style.update(process_style(content.get('style'), self))
        content['style'] = style
        pdf_content = PDFContent(content, self.fonts, x, y, width, height, self)
        return pdf_content

    def _content(
        self, content: Union[dict, 'PDFContent'], width: Number=None,
        height: Number=None, x: Number=None, y: Number=None, move: str='bottom'
    ) -> 'PDFContent':
        """Method to create and add a content box to teh current page.

        If ``content`` is a PDFContent, the method ``run`` for this instance
        will be called with the new rectangle passed to this function.
        Else, this method will try to build a new PDFContent instance with
        argument ``content`` and call method ``run`` afterwards.

        For more information about this method arguments see
        :class:`pdfme.content.PDFContent`.

        Returns:
            PDFContent: object that represents a content box.
        """
        x, y, width, height = self._position_and_size(x, y, width, height)

        if isinstance(content, PDFContent):
            pdf_content = content
            pdf_content.run(x, y, width, height)
        else:
            pdf_content = self._create_content(content, width, height, x, y)
            pdf_content.run()

        self._add_graphics([*pdf_content.fills,*pdf_content.lines])
        self._add_parts(pdf_content.parts)

        if move == 'bottom':
            self.page.y += pdf_content.current_height
        if move == 'next':
            self.page.x += width

        return pdf_content

    def content(self, content: dict) -> None:
        """Method to create and add a content box to this document. This method
        will keep adding pages to the PDF until all the contents are added to
        the document.

        Args:
            content (dict): see :class:`pdfme.content.PDFContent`.
        """
        pdf_content = self._content(
            content, x=self.page.margin_left, width=self.page.content_width
        )
        while not pdf_content.finished:
            self.add_page()
            pdf_content = self._content(pdf_content,
                self.page.content_width, self.page.content_height,
                self.page.margin_left, self.page.margin_top
            )

    def _add_graphics(self, graphics: Iterable) -> None:
        """Method to add a list of PDF graphics streams to current page.

        Args:
            graphics (list, tuple): an iterable with strings of PDF graphics
                streams.
        """
        stream = create_graphics(graphics)
        self.page.add(stream)

    def _add_parts(self, parts: Iterable) -> None:
        """Method to add a list of parts to current page.

        Each element of ``parts`` should be a dictionary with a ``type`` key,
        with value ``'text'`` for paragraph streams, or ``'image'``
        for images. The other keys in a ``text`` part are the ones obtained from
        PDFText ``result`` property, and the other keys in a ``image`` part,
        should be the same as the arguments for :meth:`pdfme.pdf.PDF.add_image`
        method.

        Args:
            parts (list, tuple): the iterable just explained.
        """
        for part in parts:
            part = part.copy()
            type_ = part.pop('type')
            if type_ == 'paragraph':
                self._add_text(**part)
            elif type_ == 'image':
                self.add_image(**part)

    def _build_pages_tree(self, page_list:list, first_level:bool=True) -> None:
        """Method to build the PDF pages tree.

        Args:
            page_list (list): a list of PDF page objects or PDFPage objects to
                be added to the PDF pages tree.
            first_level (bool, optional): whether you are calling this method
                from outside (True) or recursively from inside (False).
        """
        new_page_list = []
        count = 0
        for page in page_list:
            if first_level:
                page_size = [page.width, page.height]
                page = page.page
                page['MediaBox'] = [0, 0] + page_size

            if count % 6 == 0:
                new_page_list.append(
                    self.base.add({'Type': b'/Pages', 'Kids': [], 'Count': 0})
                )
                count += 1

            last_parent = new_page_list[-1]
            page['Parent'] = last_parent.id
            last_parent['Kids'].append(page.id)
            last_parent['Count'] += 1

        if count == 1:
            self.root['Pages'] = new_page_list[0].id
        else:
            self._build_pages_tree(new_page_list, False)

    def _build_dests_tree(
        self, keys: list, vals: list, first_level: bool=True
    ) -> None:
        """Method to build the PDF dests tree.

        Args:
            keys (list): a list of the keys of the dests to be added to the
                PDF dests tree.
            values (list): a list of the corresponding values for each key in
                ``keys`` list.
            first_level (bool, optional): whether you are calling this method
                from outside (True) or recursively from inside (False).
        """
        k = 7
        new_keys = []
        new_vals = []
        i = 0
        length = len(keys)
        obj = None
        count = 0

        while i < length:
            key = keys[i]
            val = vals[i]
            if i % k == 0:
                count += 1
                obj = self.base.add({})
                obj['Limits'] = [key if first_level else val[0], None]
                if first_level: obj['Names'] = []
                else: obj['Kids'] = []

            if first_level:
                obj['Names'].append(key)
                obj['Names'].append(val)
            else:
                obj['Kids'].append(key)

            if (i + 1) % k == 0 or (i + 1) == length:
                obj['Limits'][1] = key if first_level else val[1]
                new_keys.append(obj.id)
                new_vals.append(obj['Limits'])

            i += 1

        if count == 1:
            del obj['Limits']
            if not 'Names' in self.root: self.root['Names'] = {}
            self.root['Names']['Dests'] = obj.id
        else:
            self._build_dests_tree(new_keys, new_vals, False)

    def _build_dests(self) -> None:
        """Method to create and add the dests tree to the document.
        """
        dests = list(self.dests.keys())
        if len(dests) == 0:
            return
        dests.sort()
        self._build_dests_tree(dests, [self.dests[k] for k in dests])

    def _build_outlines_tree(
        self, outlines: list, parent: 'PDFObject', level: int
    ) -> None:
        """Method to build a PDF outline tree.

        Args:
            outlines (list): list of outlines.
            parent (PDFObject): the parent of the passed outlines.
            level (int): the level of the passed outlines.
        """
        prev = None
        count = 0
        obj = None
        for outline in outlines:
            count += 1

            obj = self.base.add({
                'Title': outline['text'],
                'Parent': parent.id,
                'Dest': outline['label']
            })

            if prev is not None:
                obj['Prev'] = prev.id
                prev['Next'] = obj.id
            else:
                parent['First'] = obj.id

            prev = obj

            children = outline.get('children', [])
            if len(children) > 0:
                count_ = self._build_outlines_tree(children, obj, level - 1)
                if level > 1:
                    count += count_
                    obj['Count'] = count_
                else:
                    obj['Count'] = -1

        parent['Last'] = obj.id
        return count

    def _build_outlines(self) -> None:
        """Method to create and add the outlines tree to the document.
        """
        if len(self.outlines) == 0:
            return

        obj = self.base.add({
            'Type': b'/Outlines'
        })
        self.root['PageMode'] = b'/UseOutlines'
        self.root['Outlines'] = obj.id
        n = self._build_outlines_tree(self.outlines, obj, self.outlines_level)
        obj['Count'] = n

    def output(self, buffer: Any) -> None:
        """Method to create the PDF file.

        Args:
            buffer (file_like): a file-like object to write the PDF file into.

        Raises:
            Exception: if this document doesn't have any pages.
        """
        if len(self.pages) == 0:
            raise Exception("pdf doesn't have any pages")
        self._build_pages_tree(self.pages)
        self._build_dests()
        self._build_outlines()
        self.base.output(buffer)

from .base import PDFBase
from .content import PDFContent
from .fonts import PDFFonts
from .image import PDFImage
from .page import PDFPage
from .parser import PDFObject
from .table import PDFTable
from .text import PDFText
from .utils import (
    create_graphics, get_page_size, get_paragraph_stream, parse_margin,
    parse_style_str, process_style, to_roman, copy
)



from collections import defaultdict
from typing import Any, Iterable, Union

from pdfme.utils import parse_range_string

STYLE_PROPS = dict(
    f='font_family', s='font_size', c='font_color', text_align='text_align',
    line_height='line_height', indent='indent'
)

PAGE_PROPS = ('page_size', 'rotate_page', 'margin')
PAGE_NUMBERING = ('page_numbering_offset', 'page_numbering_style')

class PDFDocument:
    """Class that helps to build a PDF document from a dict (``document``
    argument) describing the document contents.

    This class uses an instance of :class:`pdfme.pdf.PDF` internally to build
    the PDF document, but adds some functionalities to allow the user to
    build a PDF document from a JSONish dict, add footnotes and other
    functions explained here.

    A document is made up of sections, that can have their own page layout,
    page numbering, running sections and style.

    ``document`` dict can have the following keys:

    * ``style``: the default style of each section inside the document. A dict
      with all of the keys that a content box can have (see
      :class:`pdfme.content.PDFContent` for more information about content
      box, and for the default values of the attributes of this dict see
      :class:`pdfme.pdf.PDF`). Additional to the keys of content box style, you
      can add the following keys: ``outlines_level``, ``page_size``,
      ``rotate_page``, ``margin``, ``page_numbering_offset`` and
      ``page_numbering_style``. For more information about this page attributes
      and their default values see :class:`pdfme.pdf.PDF` definition.

    * ``formats``: a dict with the global styles of the document that can be
      used anywhere in the document. For more information about this dict
      see :class:`pdfme.pdf.PDF` definition.

    * ``running_sections``: a dict with the running sections that will be used
      by each section in the document. Each section can have, in turn, a
      ``running_section`` list, with the name of the running sections defined in
      this argument that should be included in the section. For information
      about running sections see :class:`pdfme.pdf.PDF`.
      If ``width`` key is equal to ``'left'``, it takes the value of the left
      margin, if equal to ``'right'`` it takes the value of the right margin, if
      equal to ``'full'`` it takes the value of the whole page width, and if it
      is not defined or is None it will take the value of the content width of
      the page.
      If ``height`` key is equal to ``'top'``, it takes the value of the top
      margin, if equal to ``'bottom'`` it takes the value of the bottom margin,
      if equal to ``'full'`` it takes the value of the whole page height, and if
      it is not defined or is None it will take the value the content height of
      the page.
      If ``x`` key is equal to ``'left'``, it takes the value of the left
      margin, if equal to ``'right'`` it takes the value of the whole page width
      minus the right margin, and if it is not defined or is None it will be 0.
      If ``y`` key is equal to ``'top'``, it takes the value of the top
      margin, if equal to ``'bottom'`` it takes the value of the whole page
      height minus the bottom margin, and if it is not defined or is None i
      will be 0.

    * ``per_page``: a list of dicts, each with a mandatory key ``pages``, a
      comma separated string of indexes or ranges (python style), and any of the
      following optional keys:

      * ``style``: a style dict with page related style properties (page_size,
        rotate_page, margin) that will be applied to every page in the ``pages``
        ranges.
      * ``running_sections``: a dict with optional ``exclude`` and ``include``
        lists of running sections names to be included and excluded in every
        page in the ``pages`` ranges.

    * ``sections``: an iterable with the sections of the document.

    Each section in ``sections`` iterable is a dict like the one that can be
    passed to :class:`pdfme.content.PDFContent`, so each section ends up being
    a content box. This class will add as many pages as it is needed to add
    all the contents of every section (content box) to the PDF document.
    
    Additional to the keys from a content box dict, you can
    include a ``running_sections`` list with the name of the
    running sections that you want to be included in all of the pages of the
    section. There is a special key that you can include in a section's
    ``style`` dict called ``page_numbering_reset``, that if True, resets
    the numbering of the pages.

    You can also include footnotes in any paragraph, by adding a dict with the
    key ``footnote`` with the description of the footnote as its value, to the
    list of elements of the dot key (see :class:`pdfme.text.PDFText` for more
    informarion about the structure of a paragraph and the dot key).

    Here is an example of a document dict, and how it can be used to build a
    PDF document using the helper function :func:`pdfme.document.build_pdf`.

    .. code-block:: python

        from pdfme import build_pdf

        document = {
            "style": {
                "page_size": "letter", "margin": [70, 60],
                "s": 10, "c": 0.3, "f": "Times", "text_align": "j",
                "margin_bottom": 10
            },
            "formats": {
                "link": {"c": "blue", "u": True},
                "title": {"s": 12, "b": True}
            },
            "running_sections": {
                "header": {
                    "x": "left", "y": 40, "height": "top",
                    "content": ["Document with header"]
                },
                "footer": {
                    "x": "left", "y": "bottom", "height": "bottom",
                    "style": {"text_align": "c"},
                    "content": [{".": ["Page ", {"var": "$page"}]}]
                }
            },
            "sections": [
                {
                    "running_sections": ["header", "footer"],
                    "style": {"margin": 60},
                    "content": [
                        {".": "This is a title", "style": "title"},
                        {".": [
                            "Here we include a footnote",
                            {"footnote": "Description of a footnote"},
                            ". And here we include a ",
                            {
                                ".": "link", "style": "link",
                                "uri": "https://some.url.com"
                            }
                        ]}
                    ]
                },
                {
                    "running_sections": ["footer"],
                    "style": {"rotate_page": True},
                    "content": [
                        "This is a rotated page"
                    ]
                }
            ]
        }

        with open('document.pdf', 'wb') as f:
            build_pdf(document, f)

    Args:
        document (dict): a dict like the one just described.
        context (dict, optional): a dict containing the context of the inner
            :class:`pdfme.pdf.PDF` instance.
    """
    def __init__(self, document: dict, context: dict=None) -> None:
        context = {} if context is None else context
        style = copy(document.get('style', {}))
        style_args = {
            v: style[k] for k, v in STYLE_PROPS.items() if k in style
        }

        page_args = {
            k: style[k] for k in PAGE_PROPS + PAGE_NUMBERING if k in style
        }

        self.pdf = PDF(
            outlines_level=style.get('outlines_level', 1),
            **page_args, **style_args
        )

        self.style = style

        self.pdf.context.update(context)

        self.pdf.formats = {}
        self.pdf.formats['$footnote'] = {'r': 0.5, 's': 6}
        self.pdf.formats['$footnotes'] = {'s': 10, 'c': 0}
        self.pdf.formats.update(document.get('formats', {}))

        self.running_sections = document.get('running_sections', {})

        self.per_page = []
        for range_dict in document.get('per_page', []):
            new_range_dict = copy(range_dict)
            new_range_dict['pages'] = parse_range_string(range_dict['pages'])
            self.per_page.append(new_range_dict)

        self.sections = document.get('sections', [])

        self.x = self.y = self.width = self.height = 0

        self.footnotes = []
        self._traverse_document_footnotes(self.sections)

        self.footnotes_margin = 10

    def _traverse_document_footnotes(
        self, element: Union[list, tuple, dict]
    ) -> None:
        """Method to traverse the document sections, trying to find footnotes
        dicts, to prepare them for being processed by the inner PDF instance.

        Args:
            element (list, tuple, dict): the element to be tarversed.

        Raises:
            TypeError:
        """
        if isinstance(element, (list, tuple)):
            for child in element:
                self._traverse_document_footnotes(child)
        elif isinstance(element, dict):
            if 'footnote' in element:
                element.setdefault('ids', [])
                name = '$footnote:' + str(len(self.footnotes))
                element['ids'].append(name)
                element['style'] = '$footnote'
                element['var'] = name
                self.pdf.context[name] = '0'

                footnote = element['footnote']

                if not isinstance(footnote, (dict, str, list, tuple)):
                    footnote = str(footnote)
                if isinstance(footnote, (str, list, tuple)):
                    footnote = {'.': footnote}

                if not isinstance(footnote, dict):
                    raise TypeError(
                        'footnotes must be of type dict, str, list or tuple:{}'
                        .format(footnote)
                    )

                self.footnotes.append(footnote)
            else:
                for value in element.values():
                    if isinstance(value, (list, tuple, dict)):
                        self._traverse_document_footnotes(value)

    def _set_running_sections(
        self, running_sections: Iterable, page_width: 'Number',
        page_height: 'Number', margin: dict
    ):
        """Method to set the running sections for every section in the document.

        Args:
            running_sections (list, tuple): the list of the running sections
                of the current section being added.
        """
        self.pdf.running_sections = []
        for name in running_sections:
            section = copy(self.running_sections[name])

            if section.get('width') in ['left', 'right']:
                section['width'] = margin[section.get('width')]
            if section.get('width') == 'full':
                section['width'] = page_width
            if section.get('height') in ['top', 'bottom']:
                section['height'] = margin[section.get('height')]
            if section.get('height') == 'full':
                section['height'] = page_height
            if section.get('x') == 'left':
                section['x'] = margin['left']
            if section.get('x') == 'right':
                section['x'] = page_width - margin['right']
            if section.get('y') == 'top':
                section['y'] = margin['top']
            if section.get('y') == 'bottom':
                section['y'] = page_height - margin['bottom']

            width = section.get('width', (
                page_width - margin['right'] - margin['left']
            ))
            height = section.get('height', (
                page_height - margin['top'] - margin['bottom']
            ))
            x = section.get('x', 0)
            y = section.get('y', 0)
            self.pdf.add_running_section(section, width, height, x, y)

    def run(self) -> None:
        """Method to process this document sections.
        """
        for section in self.sections:
            self._process_section(section)

    def _process_section(self, section: dict) -> None:
        """Method to process a section from this document.

        Args:
            section (dict): a dict representing the section to be processed.
        """
        section_style = copy(self.style)
        section_style.update(process_style(section.get('style', {}), self.pdf))
        
        if 'page_numbering_offset' in section_style:
            self.pdf.page_numbering_offset = section_style['page_numbering_offset']
        if 'page_numbering_style' in section_style:
            self.pdf.page_numbering_style = section_style['page_numbering_style']
        if section_style.get('page_numbering_reset', False):
            self.pdf.page_numbering_offset = -len(self.pdf.pages)

        section['style'] = section_style

        self.section = self.pdf._create_content(
            section, self.width, self.height, self.x, self.y
        )

        section_page_args = {
            k: section_style[k] for k in PAGE_PROPS if k in section_style
        }

        while True:
            page_n = len(self.pdf.pages)

            page_args = section_page_args.copy()

            running_sections = set(section.get('running_sections', []))

            for range_dict in self.per_page:
                if page_n in range_dict['pages']:
                    if 'style' in range_dict:
                        page_style = range_dict['style']
                        page_args.update({
                            k: page_style[k] for k in PAGE_PROPS
                            if k in page_style
                        })
                    if 'running_sections' in range_dict:
                        per_page_rs = range_dict['running_sections']
                        running_sections -= set(per_page_rs.get('exclude', []))
                        running_sections.update(
                            set(per_page_rs.get('include', []))
                        )

            self.pdf.setup_page(**page_args)

            page_width, page_height = self.pdf.page_width, self.pdf.page_height
            if self.pdf.rotate_page:
                page_width, page_height = page_height, page_width

            self.x = self.pdf.margin['left']
            self.width = page_width - self.pdf.margin['right'] - self.x
            self.y = page_height - self.pdf.margin['top']
            self.height = self.y - self.pdf.margin['bottom']

            self.section.setup(self.x, self.y, self.width, self.height)

            self._set_running_sections(
                running_sections, page_width, page_height, self.pdf.margin
            )

            self.pdf.add_page()

            self._add_content()

            if self.section.finished:
                break

    def _add_content(self) -> None:
        """Method to add the section contents to the current page.

        Raises:
            Exception: if the footnotes added to the page are very large.
        """

        section_state = self.section.get_state() \
            if self.section.pdf_content_part is not None else {
                'section_element_index': 0,
                'section_delayed': [],
                'children_memory': []
            }

        self.section.run(height=self.height)
        footnotes_obj = self._process_footnotes()

        if footnotes_obj is None:
            self.pdf._add_graphics([*self.section.fills,*self.section.lines])
            self.pdf._add_parts(self.section.parts)
            self.pdf.page._y -= self.section.current_height
        else:
            footnotes_height = footnotes_obj.current_height
            if footnotes_height >= self.height - self.footnotes_margin - 20:
                raise Exception(
                    "footnotes are very large and don't fit in one page"
                )
            new_height = self.height - footnotes_obj.current_height \
                - self.footnotes_margin

            if section_state is not None:
                self.section.set_state(**section_state)
                self.section.finished = False
            self.pdf._content(self.section, height=new_height)

            footnotes_obj = self._process_footnotes()

            if footnotes_obj is not None:
                self.pdf.page._y = self.pdf.margin['bottom'] + footnotes_height
                self.pdf.page.x = self.x

                x_line = round(self.pdf.page.x, 3)
                y_line = round(self.pdf.page._y + self.footnotes_margin/2, 3)
                self.pdf.page.add(' q 0 G 0.5 w {} {} m {} {} l S Q'.format(
                    x_line, y_line, x_line + 150, y_line
                ))
                self.pdf._content(footnotes_obj, height=self.height)

    def _check_footnote(self, ids: dict, page_footnotes: list) -> None:
        """Method that extract the footnotes from the ids passed as argument,
        and adds them to the ``page_footnotes`` list argument.

        Args:
            ids (dict): the ids list.
            page_footnotes (list): the list of the page footnotes to save the
                footnotes found in the ids.
        """
        for id_, rects in ids.items():
            if len(rects) == 0:
                continue
            if id_.startswith('$footnote:'):
                index = int(id_[10:])
                page_footnotes.append(self.footnotes[index])
                self.pdf.context[id_] = len(page_footnotes)

    def _check_footnotes(self, page_footnotes: list) -> None:
        """Method that loops through the current section parts, extracting the
        footnotes from each part's ids.

        Args:
            page_footnotes (list): the list of the page footnotes to save the
                footnotes found in the ids.
        """
        for part in self.section.parts:
            if part['type'] == 'paragraph':
                self._check_footnote(part['ids'], page_footnotes)

    def _get_footnotes_obj(self, page_footnotes: list) -> 'PDFContent':
        """Method to create the PDFContent object containing the footnotes of
        the current page.

        Args:
            page_footnotes (list): the list of the page footnotes

        Returns:
            PDFContent: object containing the footnotes.
        """
        content = {'style': '$footnotes', 'content': []}
        for index, footnote in enumerate(page_footnotes):
            footnote = copy(footnote)
            style = footnote.setdefault('style', {})
            style.update(dict(
                list_text=str(index + 1) + ' ', list_style='$footnote'
            ))
            content['content'].append(footnote)

        footnote_obj = self.pdf._create_content(
            content, self.width, self.height, self.x, self.y
        )
        footnote_obj.run()
        return footnote_obj

    def _process_footnotes(self) -> 'PDFContent':
        """Method to extract the footnotes from the current section parts, and
        create the PDFContent object containing the footnotes of
        the current page.

        Returns:
            PDFContent: object containing the footnotes.
        """
        page_footnotes = []
        self._check_footnotes(page_footnotes)
        if len(page_footnotes) == 0:
            return None
        return self._get_footnotes_obj(page_footnotes)

    def output(self, buffer: Any) -> None:
        """Method to create the PDF file.

        Args:
            buffer (file_like): a file-like object to write the PDF file into.
        """
        self.pdf.output(buffer)

def build_pdf(document: dict, buffer: Any, context: dict=None) -> None:
    """Function to build a PDF document using a PDFDocument instance. This is
    the easiest way to build a PDF document file in this library. For more
    information about arguments ``document``, and ``context`` see
    :class:`pdfme.document.PDFDocument`.

    Args:
        buffer (file_like): a file-like object to write the PDF file into.
    """
    doc = PDFDocument(document, context)
    doc.run()
    doc.output(buffer)

from .content import Number, PDFContent
from .pdf import PDF
from .utils import process_style, copy